# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
from .Partial_Moments import LPM, UPM


def NNS_FSD_uni(
    x: [pd.Series, np.ndarray], y: [pd.Series, np.ndarray], type_test: str = "discrete"
) -> int:
    r"""
    NNS FSD Test uni-directional

    Uni-directional test of first degree stochastic dominance using lower partial moments used in SD Efficient Set routine.

    @param x a numeric vector.
    @param y a numeric vector.
    @param type_test: ("discrete", "continuous"); \code{"discrete"} (default) selects the type of CDF.
    @return Returns (1) if \code{"X FSD Y"}, else (0).
    @author Fred Viole, OVVO Financial Systems
    @references Viole, F. and Nawrocki, D. (2016) "LPM Density Functions for the Computation of the SD Efficient Set." Journal of Mathematical Finance, 6, 105-126. \url{http://www.scirp.org/Journal/PaperInformation.aspx?PaperID=63817}.

    Viole, F. (2017) "A Note on Stochastic Dominance." \url{https://ssrn.com/abstract=3002675}.
    @examples
    set.seed(123)
    x <- rnorm(100) ; y <- rnorm(100)
    NNS.FSD.uni(x, y)
    @export
    """
    type_test = type_test.lower()

    if type_test not in ["discrete", "continuous"]:
        raise ValueError("type needs to be either discrete or continuous")
    x_min, y_min = np.min(x), np.min(y)

    if y_min > x_min:
        return 0

    Combined_sort = np.unique(np.append(x, y))
    degree = 0 if type_test == "discrete" else 1
    L_x = LPM(degree, Combined_sort, x)
    LPM_x_sort = L_x / (UPM(degree, Combined_sort, x) + L_x)
    L_y = LPM(degree, Combined_sort, y)
    LPM_y_sort = L_y / (UPM(degree, Combined_sort, y) + L_y)
    x_fsd_y = np.any(LPM_x_sort > LPM_y_sort)

    if (not x_fsd_y) and (x_min >= y_min) and (not np.equal(LPM_x_sort, LPM_y_sort).all()):
        return 1
    return 0


def NNS_SSD_uni(x: [pd.Series, np.ndarray], y: [pd.Series, np.ndarray]) -> int:
    r"""
    NNS SSD Test uni-directional

    Uni-directional test of second degree stochastic dominance using lower partial moments used in SD Efficient Set routine.
    @param x a numeric vector.
    @param y a numeric vector.
    @return Returns (1) if \code{"X SSD Y"}, else (0).
    @author Fred Viole, OVVO Financial Systems
    @references Viole, F. and Nawrocki, D. (2016) "LPM Density Functions for the Computation of the SD Efficient Set."
        Journal of Mathematical Finance, 6, 105-126. \url{http://www.scirp.org/Journal/PaperInformation.aspx?PaperID=63817}.
    @examples
    set.seed(123)
    x <- rnorm(100) ; y <- rnorm(100)
    NNS.SSD.uni(x, y)
    @export
    """
    x_min, y_min = np.min(x), np.min(y)
    if y_min > x_min:
        return 0
    x_mean, y_mean = np.mean(x), np.mean(y)
    if y_mean > x_mean:
        return 0
    Combined_sort = np.unique(np.append(x, y))
    LPM_x_sort = LPM(1, Combined_sort, x)
    LPM_y_sort = LPM(1, Combined_sort, y)
    x_ssd_y = np.any(LPM_x_sort > LPM_y_sort)
    if (not x_ssd_y) and (x_min >= y_min) and (not np.equal(LPM_x_sort, LPM_y_sort).all()):
        return 1
    return 0


def NNS_TSD_uni(x: [pd.Series, np.ndarray], y: [pd.Series, np.ndarray]) -> int:
    r"""
    NNS TSD Test uni-directional

    Uni-directional test of third degree stochastic dominance using lower partial moments used in SD Efficient Set routine.
    @param x a numeric vector.
    @param y a numeric vector.
    @return Returns (1) if \code{"X TSD Y"}, else (0).
    @author Fred Viole, OVVO Financial Systems
    @references Viole, F. and Nawrocki, D. (2016) "LPM Density Functions for the Computation of the SD Efficient Set."
        Journal of Mathematical Finance, 6, 105-126. \url{http://www.scirp.org/Journal/PaperInformation.aspx?PaperID=63817}.
    @examples
    set.seed(123)
    x <- rnorm(100) ; y <- rnorm(100)
    NNS.TSD.uni(x, y)
    @export
    """
    x_min, y_min = np.min(x), np.min(y)
    if y_min > x_min:
        return 0
    x_mean, y_mean = np.mean(x), np.mean(y)
    if y_mean > x_mean:
        return 0
    Combined_sort = np.unique(np.append(x, y))
    LPM_x_sort = LPM(2, Combined_sort, x)
    LPM_y_sort = LPM(2, Combined_sort, y)
    x_tsd_y = np.any(LPM_x_sort > LPM_y_sort)

    if (not x_tsd_y) and (x_min >= y_min) and (not np.equal(LPM_x_sort, LPM_y_sort).all()):
        return 1
    return 0


__all__ = ["NNS_FSD_uni", "NNS_SSD_uni", "NNS_TSD_uni"]



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd

def NNS_dep_matrix(x: [np.ndarray, pd.Series], order=None, degree=None, asym: bool=False):
    if len(x.shape) == 1:
        raise Exception("supply both 'x' and 'y' or a matrix-like 'x'")
    n = x.shape[1]
    if isinstance(x, pd.Series):
        x = x.to_frame()
    if isinstance(x, np.ndarray) and len(x.shape) == 1:
        x = x.reshape(-1, 1)  # series to matrix like

    if(x.shape[0] < 20 ):
        order=2

    raw_rhos_lower = {}
    raw_deps_lower = {}
    raw_both_lower = {}
    raw_rhos_upper = {}
    raw_deps_upper = {}
    raw_both_upper = {}

    for i in range(n):
        raw_both_lower[i] = sapply((i + 1) : n, function(b) NNS_dep(x[ , i], x[ , b], print.map = FALSE, asym = asym))
        raw_both_upper[i] = sapply((i + 1) : n, function(b) NNS_dep(x[ , b], x[ , i], print.map = FALSE, asym = asym))
        raw_rhos_upper[i] = unlist(raw_both_upper[i][row_names(raw_both_upper[i])=="Correlation"])
        raw_deps_upper[i] = unlist(raw_both_upper[i][row_names(raw_both_upper[i])=="Dependence"])
        raw_rhos_lower[i] = unlist(raw_both_lower[i][row_names(raw_both_lower[i])=="Correlation"])
        raw_deps_lower[i] = unlist(raw_both_lower[i][row_names(raw_both_lower[i])=="Dependence"])

    rhos = matrix(, n, n)
    deps = matrix(0, n, n)
    if not asym:
        rhos[lower.tri(rhos, diag = FALSE)] = (unlist(raw.rhos_upper) + unlist(raw.rhos_lower)) / 2
        deps[lower.tri(deps, diag = FALSE)] = (unlist(raw.deps_upper) + unlist(raw.deps_lower)) / 2
        rhos = pmax(rhos, t(rhos), na.rm = TRUE)
        deps = pmax(deps, t(deps), na.rm = TRUE)
    else:
        rhos[lower.tri(rhos, diag = FALSE)] = unlist(raw.rhos_lower)
        deps[lower.tri(deps, diag = FALSE)] = unlist(raw.deps_lower)

        rhos_upper = matrix(0, n, n)
        deps_upper = matrix(0, n, n)

        rhos[is.na(rhos)] = 0
        deps[is.na(deps)] = 0

        rhos_upper[lower.tri(rhos_upper, diag=FALSE)] = unlist(raw.rhos_upper)
        rhos_upper = t(rhos_upper)

        deps_upper[lower.tri(deps_upper, diag=FALSE)] = unlist(raw.deps_upper)
        deps_upper = t(deps_upper)

        rhos = rhos + rhos_upper
        deps = deps + deps_upper

    diag(rhos) = 1
    diag(deps) = 1

    colnames(rhos) = colnames(x)
    colnames(deps) = colnames(x)
    rownames(rhos) = colnames(x)
    rownames(deps) = colnames(x)

    return {
        "Correlation": rhos,
        "Dependence": deps
    }

__all__ = [
    "NNS_dep_matrix"
]


# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
import numba


def pd_fill_diagonal(df_matrix: pd.DataFrame, value) -> None:
    n = min(df_matrix.shape[0], df_matrix.shape[1])
    df_matrix.values[tuple([np.arange(n)] * 2)] = value


@numba.jit(parallel=True, nopython=True)
def numba_LPM(degree: [int, float], target: np.ndarray, variable: np.ndarray) -> np.ndarray:
    ret = np.zeros(shape=(target.shape[0]), dtype=np.float32)
    for i in numba.prange(target.shape[0]):
        # ugly implementation, but working
        for ll in range(variable.shape[0]):
            if variable[ll] <= target[i]:
                ret[i] += (target[i] - variable[ll]) ** degree
        ret[i] /= variable.shape[0]

        # TODO:  variable[variable <= target[i]] didn't work at numba
        # ret[i] += (
        #    (target[i] - (variable[variable <= target[i]])) ** degree
        # ).sum() / variable.shape[0]
    return ret


def LPM(
    degree: [int, float],
    target: [int, float, str, None, pd.Series, np.ndarray, list],
    variable: [pd.Series, np.ndarray, list],
) -> [float, np.ndarray]:
    r"""
    Lower Partial Moment

    This function generates a univariate lower partial moment for any degree or target.
    @param degree integer; \code{(degree = 0)} is frequency, \code{(degree = 1)} is area.
    @param target numeric; Typically set to mean, but does not have to be. (Vectorized)
    @param variable a numeric vector.
    @return LPM of variable
    @author Fred Viole, OVVO Financial Systems
    @references Viole, F. and Nawrocki, D. (2013) "Nonlinear Nonparametric Statistics: Using Partial Moments"
    \url{https://www.amazon.com/dp/1490523995}
    @examples
    set.seed(123)
    x <- rnorm(100)
    LPM(0, mean(x), x)
    @export
    """

    if target is None:
        target = np.mean(variable)
    if isinstance(target, str):  # "mean"
        target = getattr(np, target)(variable)
    if isinstance(target, list):
        target = np.array(target)
    if isinstance(variable, list):
        variable = np.array(variable)
    if degree == 0:
        if isinstance(target, (np.ndarray, pd.Series, list)):
            return np.array([np.mean(variable <= i) for i in target])
        return np.mean(variable <= target)
    if isinstance(target, (np.ndarray, list)):
        return numba_LPM(
            degree=degree,
            target=target,
            variable=variable if not hasattr(variable, "values") else variable.values,
        )
    elif isinstance(target, pd.Series):
        return numba_LPM(
            degree=degree,
            target=target.values,
            variable=variable if not hasattr(variable, "values") else variable.values,
        )
    return numba_LPM(
        degree=degree,
        target=np.array([target]),
        variable=variable if not hasattr(variable, "values") else variable.values,
    )[0]


@numba.jit(parallel=True, nopython=True)
def numba_UPM(degree: [int, float], target: np.ndarray, variable: np.ndarray) -> np.ndarray:
    ret = np.zeros(shape=(target.shape[0]), dtype=np.float32)
    for i in numba.prange(target.shape[0]):
        # ugly implementation, but working
        for ll in range(variable.shape[0]):
            if variable[ll] > target[i]:
                ret[i] += (variable[ll] - target[i]) ** degree
        ret[i] /= variable.shape[0]
        # TODO:  variable[variable > target[i]] didn't work at numba
        # ret[i] += (
        #    ((variable[variable > target[i]]) - target[i]) ** degree
        # ).sum() / variable.shape[0]
    return ret


def UPM(
    degree: [int, float],
    target: [int, float, str, None, pd.Series, np.ndarray, list],
    variable: [pd.Series, np.ndarray, list],
) -> [float, np.ndarray]:
    r"""
    Upper Partial Moment

    This function generates a univariate upper partial moment for any degree or target.
    @param degree integer; \code{(degree = 0)} is frequency, \code{(degree = 1)} is area.
    @param target numeric; Typically set to mean, but does not have to be. (Vectorized)
    @param variable a numeric vector.
    @return UPM of variable
    @author Fred Viole, OVVO Financial Systems
    @references Viole, F. and Nawrocki, D. (2013) "Nonlinear Nonparametric Statistics: Using Partial Moments"
    \url{https://www.amazon.com/dp/1490523995}
    @examples
    set.seed(123)
    x <- rnorm(100)
    UPM(0, mean(x), x)
    @export
    """
    if target is None:
        target = np.mean(variable)
    if isinstance(target, str):  # "mean"
        target = getattr(np, target)(variable)
    if isinstance(target, list):
        target = np.array(target)
    if isinstance(variable, list):
        variable = np.array(variable)
    if degree == 0:
        if isinstance(target, (np.ndarray, pd.Series, list)):
            return np.array([np.mean(variable > i) for i in target])
        return np.mean(variable > target)
    if isinstance(target, (np.ndarray, list)):
        return numba_UPM(
            degree=degree,
            target=target,
            variable=variable if not hasattr(variable, "values") else variable.values,
        )
    elif isinstance(target, pd.Series):
        return numba_UPM(
            degree=degree,
            target=target.values,
            variable=variable if not hasattr(variable, "values") else variable.values,
        )
    return numba_UPM(
        degree=degree,
        target=np.array([target]),
        variable=variable if not hasattr(variable, "values") else variable.values,
    )[0]


def _Co_UPM(
    degree_x: [float, int],
    degree_y: [float, int],
    x: [pd.Series, np.ndarray, list],
    y: [pd.Series, np.ndarray, list],
    target_x: [int, float, str, None] = None,
    target_y: [int, float, str, None] = None,
) -> float:
    if target_x is None:
        target_x = np.mean(x)
    if target_y is None:
        target_y = np.mean(y)
    if isinstance(target_x, str):  # "mean"
        target_x = getattr(np, target_x)(x)
    if isinstance(target_y, str):  # "mean"
        target_y = getattr(np, target_y)(y)

    z = pd.DataFrame({"x": x, "y": y})
    z["x"] = z["x"] - target_x
    z["y"] = z["y"] - target_y

    z.loc[z["x"] <= 0, "x"] = np.nan
    z.loc[z["y"] <= 0, "y"] = np.nan

    z.dropna(inplace=True)

    return (z["x"] ** degree_x).dot(z["y"] ** degree_y) / len(x)


# Co.UPM <- Vectorize(Co.UPM, vectorize.args = c('target.x', 'target.y'))
_vec_Co_UPM = np.vectorize(_Co_UPM, excluded=["degree_x", "degree_y", "x", "y"])


def Co_UPM(
    degree_x: [float, int],
    degree_y: [float, int],
    x: [pd.Series, np.ndarray, list],
    y: [pd.Series, np.ndarray, list],
    target_x: [int, float, str, None, pd.Series, np.ndarray, list] = None,
    target_y: [int, float, str, None, pd.Series, np.ndarray, list] = None,
) -> [float, np.ndarray]:
    r"""
    Co-Upper Partial Moment
    (Upper Right Quadrant 1)

    This function generates a co-upper partial moment between two equal length variables for any degree or target.
    @param degree_x integer; Degree for variable X.  \code{(degree.x = 0)} is frequency, \code{(degree.x = 1)} is area.
    @param degree_y integer; Degree for variable Y.  \code{(degree.y = 0)} is frequency, \code{(degree.y = 1)} is area.
    @param x a numeric vector.
    @param y a numeric vector of equal length to \code{x}.
    @param target_x numeric; Typically the mean of Variable X for classical statistics equivalences, but does not have to be. (Vectorized)
    @param target_y numeric; Typically the mean of Variable Y for classical statistics equivalences, but does not have to be. (Vectorized)
    @return Co-UPM of two variables
    @author Fred Viole, OVVO Financial Systems
    @references Viole, F. and Nawrocki, D. (2013) "Nonlinear Nonparametric Statistics: Using Partial Moments"
    \url{https://www.amazon.com/dp/1490523995}
    @examples
    set.seed(123)
    x <- rnorm(100) ; y <- rnorm(100)
    Co.UPM(0, 0, x, y, mean(x), mean(y))
    @export
    """
    func = _Co_UPM
    if isinstance(target_x, list):
        target_x = np.array(target_x)
    if isinstance(target_y, list):
        target_y = np.array(target_y)
    if isinstance(target_y, (np.ndarray, pd.Series)) or isinstance(
        target_x, (np.ndarray, pd.Series)
    ):
        func = _vec_Co_UPM
    return func(
        degree_x=degree_x,
        degree_y=degree_y,
        x=x,
        y=y,
        target_x=target_x,
        target_y=target_y,
    )


def _Co_LPM(
    degree_x: [float, int],
    degree_y: [float, int],
    x: [pd.Series, np.ndarray, list],
    y: [pd.Series, np.ndarray, list],
    target_x: [int, float, str, None] = None,
    target_y: [int, float, str, None] = None,
) -> float:
    if target_x is None:
        target_x = np.mean(x)
    if target_y is None:
        target_y = np.mean(y)
    if isinstance(target_x, str):  # "mean"
        target_x = getattr(np, target_x)(x)
    if isinstance(target_y, str):  # "mean"
        target_y = getattr(np, target_y)(y)

    z = pd.DataFrame({"x": x, "y": y})
    # z <- t(c(target.x, target.y) - t(z))
    z["x"] = target_x - z["x"]
    z["y"] = target_y - z["y"]

    z.loc[z["x"] <= 0, "x"] = np.nan
    z.loc[z["y"] <= 0, "y"] = np.nan

    z.dropna(inplace=True)

    return (z["x"] ** degree_x).dot(z["y"] ** degree_y) / len(x)


# Co.LPM <- Vectorize(Co.LPM, vectorize.args = c('target.x', 'target.y'))
_vec_Co_LPM = np.vectorize(_Co_LPM, excluded=["degree_x", "degree_y", "x", "y"])


def Co_LPM(
    degree_x: [float, int],
    degree_y: [float, int],
    x: [pd.Series, np.ndarray, list],
    y: [pd.Series, np.ndarray, list],
    target_x: [int, float, str, None, np.ndarray, pd.Series, list] = None,
    target_y: [int, float, str, None, np.ndarray, pd.Series, list] = None,
) -> [float, np.ndarray]:
    r"""
    Co-Lower Partial Moment
    (Lower Left Quadrant 4)

    This function generates a co-lower partial moment for between two equal length variables for any degree or target.
    @param degree_x integer; Degree for variable X.  \code{(degree.x = 0)} is frequency, \code{(degree.x = 1)} is area.
    @param degree_y integer; Degree for variable Y.  \code{(degree.y = 0)} is frequency, \code{(degree.y = 1)} is area.
    @param x a numeric vector.
    @param y a numeric vector of equal length to \code{x}.
    @param target_x numeric; Typically the mean of Variable X for classical statistics equivalences, but does not have to be. (Vectorized)
    @param target_y numeric; Typically the mean of Variable Y for classical statistics equivalences, but does not have to be. (Vectorized)
    @return Co-LPM of two variables
    @author Fred Viole, OVVO Financial Systems
    @references Viole, F. and Nawrocki, D. (2013) "Nonlinear Nonparametric Statistics: Using Partial Moments"
    \url{https://www.amazon.com/dp/1490523995}
    @examples
    set.seed(123)
    x <- rnorm(100) ; y <- rnorm(100)
    Co.LPM(0, 0, x, y, mean(x), mean(y))
    @export
    """
    func = _Co_LPM
    if isinstance(target_x, list):
        target_x = np.array(target_x)
    if isinstance(target_y, list):
        target_y = np.array(target_y)
    if isinstance(target_y, (np.ndarray, pd.Series)) or isinstance(
        target_x, (np.ndarray, pd.Series)
    ):
        func = _vec_Co_LPM
    return func(
        degree_x=degree_x,
        degree_y=degree_y,
        x=x,
        y=y,
        target_x=target_x,
        target_y=target_y,
    )


def _D_LPM(
    degree_x: [float, int],
    degree_y: [float, int],
    x: [pd.Series, np.ndarray, list],
    y: [pd.Series, np.ndarray, list],
    target_x: [int, float, str, None] = None,
    target_y: [int, float, str, None] = None,
) -> float:
    if target_x is None:
        target_x = np.mean(x)
    if target_y is None:
        target_y = np.mean(y)
    if isinstance(target_x, str):  # "mean"
        target_x = getattr(np, target_x)(x)
    if isinstance(target_y, str):  # "mean"
        target_y = getattr(np, target_y)(y)

    z = pd.DataFrame({"x": x, "y": y})
    #   z[,1] <- z[,1] - target.x
    #   z[,2] <- target.y - z[,2]
    z["x"] = z["x"] - target_x
    z["y"] = target_y - z["y"]

    z.loc[z["x"] <= 0, "x"] = np.nan
    z.loc[z["y"] <= 0, "y"] = np.nan

    z.dropna(inplace=True)

    return (z["x"] ** degree_x).dot(z["y"] ** degree_y) / len(x)


# D.LPM <- Vectorize(D.LPM, vectorize.args = c('target.x', 'target.y'))
_vec_D_LPM = np.vectorize(_D_LPM, excluded=["degree_x", "degree_y", "x", "y"])


def D_LPM(
    degree_x: [float, int],
    degree_y: [float, int],
    x: [pd.Series, np.ndarray, list],
    y: [pd.Series, np.ndarray, list],
    target_x: [int, float, str, None, pd.Series, np.ndarray, list] = None,
    target_y: [int, float, str, None, pd.Series, np.ndarray, list] = None,
) -> float:
    r"""
    Divergent-Lower Partial Moment
    (Lower Right Quadrant 3)

    This function generates a divergent lower partial moment between two equal length variables for any degree or target.
    @param degree_x integer; Degree for variable X.  \code{(degree.x = 0)} is frequency, \code{(degree.x = 1)} is area.
    @param degree_y integer; Degree for variable Y.  \code{(degree.y = 0)} is frequency, \code{(degree.y = 1)} is area.
    @param x a numeric vector.
    @param y a numeric vector of equal length to \code{x}.
    @param target_x numeric; Typically the mean of Variable X for classical statistics equivalences, but does not have to be. (Vectorized)
    @param target_y numeric; Typically the mean of Variable Y for classical statistics equivalences, but does not have to be. (Vectorized)
    @return Divergent LPM of two variables
    @author Fred Viole, OVVO Financial Systems
    @references Viole, F. and Nawrocki, D. (2013) "Nonlinear Nonparametric Statistics: Using Partial Moments"
    \url{https://www.amazon.com/dp/1490523995}
    @examples
    set.seed(123)
    x <- rnorm(100) ; y <- rnorm(100)
    D.LPM(0, 0, x, y, mean(x), mean(y))
    @export
    """
    func = _D_LPM
    if isinstance(target_x, list):
        target_x = np.array(target_x)
    if isinstance(target_y, list):
        target_y = np.array(target_y)
    if isinstance(target_y, (np.ndarray, pd.Series)) or isinstance(
        target_x, (np.ndarray, pd.Series)
    ):
        func = _vec_D_LPM
    return func(
        degree_x=degree_x,
        degree_y=degree_y,
        x=x,
        y=y,
        target_x=target_x,
        target_y=target_y,
    )


def _D_UPM(
    degree_x: [int, float],
    degree_y: [int, float],
    x: [pd.Series, np.ndarray, list],
    y: [pd.Series, np.ndarray, list],
    target_x: [int, float, str, None] = None,
    target_y: [int, float, str, None] = None,
) -> float:
    if target_x is None:
        target_x = np.mean(x)
    if target_y is None:
        target_y = np.mean(y)
    if isinstance(target_x, str):  # "mean"
        target_x = getattr(np, target_x)(x)
    if isinstance(target_y, str):  # "mean"
        target_y = getattr(np, target_y)(y)
    z = pd.DataFrame({"x": x, "y": y})
    # z[,1] <- target.x - z[,1]
    # z[,2] <- z[,2] - target.y
    z["x"] = target_x - z["x"]
    z["y"] = z["y"] - target_y

    z.loc[z["x"] <= 0, "x"] = np.nan
    z.loc[z["y"] <= 0, "y"] = np.nan

    z.dropna(inplace=True)

    return (z["x"] ** degree_x).dot(z["y"] ** degree_y) / len(x)


# D.UPM <- Vectorize(D.UPM, vectorize.args = c('target.x', 'target.y'))
_vec_D_UPM = np.vectorize(_D_UPM, excluded=["degree_x", "degree_y", "x", "y"])


def D_UPM(
    degree_x: [int, float],
    degree_y: [int, float],
    x: [pd.Series, np.ndarray, list],
    y: [pd.Series, np.ndarray, list],
    target_x: [int, float, str, None, pd.Series, np.ndarray, list] = None,
    target_y: [int, float, str, None, pd.Series, np.ndarray, list] = None,
) -> float:
    r"""
    Divergent-Upper Partial Moment
    (Upper Left Quadrant 2)

    This function generates a divergent upper partial moment between two equal length variables for any degree or target.
    @param degree_x integer; Degree for variable X.  \code{(degree.x = 0)} is frequency, \code{(degree.x = 1)} is area.
    @param degree_y integer; Degree for variable Y.  \code{(degree.y = 0)} is frequency, \code{(degree.y = 1)} is area.
    @param x a numeric vector.
    @param y a numeric vector of equal length to \code{x}.
    @param target_x numeric; Typically the mean of Variable X for classical statistics equivalences, but does not have to be. (Vectorized)
    @param target_y numeric; Typically the mean of Variable Y for classical statistics equivalences, but does not have to be. (Vectorized)
    @return Divergent UPM of two variables
    @author Fred Viole, OVVO Financial Systems
    @references Viole, F. and Nawrocki, D. (2013) "Nonlinear Nonparametric Statistics: Using Partial Moments"
    \url{https://www.amazon.com/dp/1490523995}
    @examples
    set.seed(123)
    x <- rnorm(100) ; y <- rnorm(100)
    D.UPM(0, 0, x, y, mean(x), mean(y))
    @export
    """
    func = _D_UPM
    if isinstance(target_x, list):
        target_x = np.array(target_x)
    if isinstance(target_y, list):
        target_y = np.array(target_y)
    if isinstance(target_y, (np.ndarray, pd.Series)) or isinstance(
        target_x, (np.ndarray, pd.Series)
    ):
        func = _vec_D_UPM
    return func(
        degree_x=degree_x,
        degree_y=degree_y,
        x=x,
        y=y,
        target_x=target_x,
        target_y=target_y,
    )


def PM_matrix(
    LPM_degree: [int, float],
    UPM_degree: [int, float],
    target: [str, dict, list, float, int, pd.Series, np.array, list] = "mean",
    variable: [pd.Series, pd.DataFrame, np.ndarray, None, list] = None,
    pop_adj: bool = False,
) -> dict:
    r"""
    Partial Moment Matrix


    This function generates a co-partial moment matrix for the specified co-partial moment.
    @param LPM_degree integer; Degree for \code{variable} below \code{target} deviations.  \code{(degree = 0)} is frequency, \code{(degree = 1)} is area.
    @param UPM_degree integer; Degree for \code{variable} above \code{target} deviations.  \code{(degree = 0)} is frequency, \code{(degree = 1)} is area.
    @param target numeric; Typically the mean of Variable X for classical statistics equivalences, but does not have to be. (Vectorized)  \code{(target = "mean")} (default) will set the target as the mean of every variable.
    @param variable a numeric matrix or data.frame.
    @param pop_adj logical; \code{FALSE} (default) Adjusts the sample co-partial moment matrices for population statistics.
    @return Matrix of partial moment quadrant values (CUPM, DUPM, DLPM, CLPM), and overall covariance matrix.  Uncalled quadrants will return a matrix of zeros.
    @note For divergent asymmetical \code{"D.LPM" and "D.UPM"} matrices, matrix is \code{D.LPM(column,row,...)}.
    @author Fred Viole, OVVO Financial Systems
    @references Viole, F. and Nawrocki, D. (2013) "Nonlinear Nonparametric Statistics: Using Partial Moments"
    \url{https://www.amazon.com/dp/1490523995}
    @references Viole, F. (2017) "Bayes' Theorem From Partial Moments"
    \url{https://ssrn.com/abstract=3457377}
    @examples
    set.seed(123)
    x <- rnorm(100) ; y <- rnorm(100) ; z <- rnorm(100)
    A <- cbind(x,y,z)
    PM.matrix(LPM.degree = 1, UPM.degree = 1, target = "mean", variable = A)

    ## Use of vectorized numeric targets (target_x, target_y, target_z)
    PM.matrix(LPM.degree = 1, UPM.degree = 1, target = c(0, 0.15, .25), variable = A)

    ## Calling Individual Partial Moment Quadrants
    cov.mtx <- PM.matrix(LPM.degree = 1, UPM.degree = 1, target = "mean", variable = A)
    cov.mtx$cupm

    ## Full covariance matrix
    cov.mtx$cov.matrix
    @export
    """

    if variable is None:
        return {
            "cupm": None,
            "dupm": None,
            "dlpm": None,
            "clpm": None,
            "cov.matrix": None,
        }
    if isinstance(variable, list):
        variable = np.column_stack(variable)
    if isinstance(target, list):
        target = np.array(target)
    if isinstance(variable, pd.Series):
        variable = variable.to_frame()
    if isinstance(variable, np.ndarray) and len(variable.shape) == 1:
        variable = variable.reshape(-1, 1)  # series to matrix like
    assert isinstance(
        variable, (pd.DataFrame, np.ndarray, np.matrix)
    ), "supply a matrix-like (pd.DataFrame, np.ndarray, np.matrix) 'variable'"

    n = variable.shape[1]
    variable_columns = (
        variable.columns if isinstance(variable, pd.DataFrame) else list(range(variable.shape[1]))
    )

    # target dict
    if isinstance(target, (list, pd.Series, np.ndarray)):
        target = {i: v for i, v in enumerate(target)}
    elif isinstance(target, str):
        # mean / median / mode
        if isinstance(variable, pd.DataFrame):
            target = {i: getattr(np, target)(variable.values[:, i]) for i in range(n)}
        else:
            target = {i: getattr(np, target)(variable[:, i]) for i in range(n)}
    elif isinstance(target, (int, float)):
        target = {i: target for i in range(n)}
    # Partial moments lists
    clpms, cupms, dlpms, dupms = [], [], [], []
    for cur_var in range(n):
        clpms.append([])
        cupms.append([])
        dlpms.append([])
        dupms.append([])
        # sapply(X, FUN, ..., simplify = TRUE, USE.NAMES = TRUE)
        # clpms[[i]] <- sapply(1 : n, function(b) Co.LPM(x = variable[ , i], y = variable[ , b], degree.x = LPM.degree, degree.y = LPM.degree, target.x = target[i], target.y = target[b]))
        for cur_var2 in range(n):
            if isinstance(variable, pd.DataFrame):
                x = variable.values[:, cur_var]
                y = variable.values[:, cur_var2]
            else:
                x = variable[:, cur_var]
                y = variable[:, cur_var2]
            clpms[cur_var].append(
                Co_LPM(
                    x=x,
                    y=y,
                    degree_x=LPM_degree,
                    degree_y=LPM_degree,
                    target_x=target[cur_var],
                    target_y=target[cur_var2],
                )
            )
            cupms[cur_var].append(
                Co_UPM(
                    x=x,
                    y=y,
                    degree_x=UPM_degree,
                    degree_y=UPM_degree,
                    target_x=target[cur_var],
                    target_y=target[cur_var2],
                )
            )
            #            dlpms[[i]] <- sapply(1 : n, function(b)
            #            D.LPM(x = variable[ , i], y = variable[ , b], degree.x = UPM.degree, degree.y = LPM.degree, target.x = target[i], target.y = target[b]))
            if cur_var == cur_var2:
                dlpms[cur_var].append(0.0)
                dupms[cur_var].append(0.0)
            else:
                dlpms[cur_var].append(
                    D_LPM(
                        x=x,
                        y=y,
                        degree_x=UPM_degree,
                        degree_y=LPM_degree,
                        target_x=target[cur_var],
                        target_y=target[cur_var2],
                    )
                )
                dupms[cur_var].append(
                    D_UPM(
                        x=x,
                        y=y,
                        degree_x=LPM_degree,
                        degree_y=UPM_degree,
                        target_x=target[cur_var],
                        target_y=target[cur_var2],
                    )
                )

    # clpm.matrix <- matrix(unlist(clpms), n, n)
    # colnames(clpm.matrix) <- colnames(variable)
    # rownames(clpm.matrix) <- colnames(variable)
    clpm_matrix = pd.DataFrame(clpms, index=variable_columns, columns=variable_columns).T

    # cupm.matrix <- matrix(unlist(cupms), n, n)
    # colnames(cupm.matrix) <- colnames(variable)
    # rownames(cupm.matrix) <- colnames(variable)
    cupm_matrix = pd.DataFrame(cupms, index=variable_columns, columns=variable_columns).T

    # dlpm.matrix <- matrix(unlist(dlpms), n, n)
    # diag(dlpm.matrix) <- 0
    # colnames(dlpm.matrix) <- colnames(variable)
    # rownames(dlpm.matrix) <- colnames(variable)
    dlpm_matrix = pd.DataFrame(dlpms, index=variable_columns, columns=variable_columns).T
    # pd_fill_diagonal(dlpm_matrix, 0.0)

    # dupm.matrix <- matrix(unlist(dupms), n, n)
    # diag(dupm.matrix) <- 0
    # colnames(dupm.matrix) <- colnames(variable)
    # rownames(dupm.matrix) <- colnames(variable)
    dupm_matrix = pd.DataFrame(dupms, index=variable_columns, columns=variable_columns).T
    # pd_fill_diagonal(dupm_matrix, 0.0)

    if pop_adj:
        # adjustment <- length(variable[ , 1]) / (length(variable[ , 1]) - 1)
        adjustment = variable.shape[1] / (variable.shape[1] - 1)
        clpm_matrix *= adjustment
        cupm_matrix *= adjustment
        dlpm_matrix *= adjustment
        dupm_matrix *= adjustment

    # cov.matrix <- cupm.matrix + clpm.matrix - dupm.matrix - dlpm.matrix
    cov_matrix = cupm_matrix + clpm_matrix - dupm_matrix - dlpm_matrix

    return {
        "cupm": cupm_matrix,
        "dupm": dupm_matrix,
        "dlpm": dlpm_matrix,
        "clpm": clpm_matrix,
        "cov.matrix": cov_matrix,
    }


def LPM_ratio(
    degree: [int, float],
    target: [int, float, str, None, pd.Series, np.ndarray, list],
    variable: [pd.Series, np.ndarray, list],
) -> [float, np.ndarray]:
    r"""
    Lower Partial Moment RATIO

    This function generates a standardized univariate lower partial moment for any degree or target.
    @param degree integer; \code{(degree = 0)} is frequency, \code{(degree = 1)} is area.
    @param target numeric; Typically set to mean, but does not have to be. (Vectorized)
    @param variable a numeric vector.
    @return Standardized LPM of variable
    @author Fred Viole, OVVO Financial Systems
    @references Viole, F. and Nawrocki, D. (2013) "Nonlinear Nonparametric Statistics: Using Partial Moments"
    \url{https://www.amazon.com/dp/1490523995}
    @references Viole, F. (2017) "Continuous CDFs and ANOVA with NNS"
    \url{https://ssrn.com/abstract=3007373}
    @examples
    set.seed(123)
    x <- rnorm(100)
    LPM.ratio(0, mean(x), x)

    \dontrun{
    ## Empirical CDF (degree = 0)
    lpm_cdf <- LPM.ratio(0, sort(x), x)
    plot(sort(x), lpm_cdf)

    ## Continuous CDF (degree = 1)
    lpm_cdf_1 <- LPM.ratio(1, sort(x), x)
    plot(sort(x), lpm_cdf_1)

    ## Joint CDF
    x <- rnorm(5000) ; y <- rnorm(5000)
    plot3d(x, y, Co.LPM(0, 0, sort(x), sort(y), x, y), col = "blue", xlab = "X", ylab = "Y",
    zlab = "Probability", box = FALSE)
    }
    @export
    """
    lpm = LPM(degree=degree, target=target, variable=variable)
    if degree > 0:
        area = lpm + UPM(degree=degree, target=target, variable=variable)
    else:
        area = 1
    return lpm / area


def UPM_ratio(
    degree: [int, float],
    target: [int, float, str, None, list, np.ndarray, pd.Series],
    variable: [pd.Series, np.ndarray, list],
) -> [float, np.ndarray]:
    r"""
    Upper Partial Moment RATIO

    This function generates a standardized univariate upper partial moment for any degree or target.
    @param degree integer; \code{(degree = 0)} is frequency, \code{(degree = 1)} is area.
    @param target numeric; Typically set to mean, but does not have to be. (Vectorized)
    @param variable a numeric vector.
    @return Standardized UPM of variable
    @author Fred Viole, OVVO Financial Systems
    @references Viole, F. and Nawrocki, D. (2013) "Nonlinear Nonparametric Statistics: Using Partial Moments"
    \url{https://www.amazon.com/dp/1490523995}
    @examples
    set.seed(123)
    x <- rnorm(100)
    UPM.ratio(0, mean(x), x)

    ## Joint Upper CDF
    \dontrun{
    x <- rnorm(5000) ; y <- rnorm(5000)
    plot3d(x, y, Co.UPM(0, 0, sort(x), sort(y), x, y), col = "blue", xlab = "X", ylab = "Y",
    zlab = "Probability", box = FALSE)
    }
    @export
    """
    upm = UPM(degree, target, variable)
    if degree > 0:
        area = LPM(degree, target, variable) + upm
    else:
        area = 1
    return upm / area


def NNS_PDF(
    variable: pd.Series,
    degree: [int, float] = 1,
    target: [float, int, str, None] = None,
    bins: [int, None] = None,
    plot: bool = True,
) -> pd.DataFrame:
    # TODO: CONVERT/TEST
    if target is None:
        target = variable.sort_values()
    # d/dx approximation
    if bins is None:
        # python:
        # https://numpy.org/doc/stable/reference/generated/numpy.histogram.html
        # https://stackoverflow.com/questions/9141732/how-does-numpy-histogram-work
        # https://stackoverflow.com/questions/33967513/creating-density-estimate-in-numpy
        # https://jakevdp.github.io/blog/2013/12/01/kernel-density-estimation/
        # https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gaussian_kde.html

        #
        ## Default S3 method:
        # density(x, bw = "nrd0", adjust = 1,
        #        kernel = c("gaussian", "epanechnikov", "rectangular",
        #                   "triangular", "biweight",
        #                   "cosine", "optcosine"),
        #        weights = NULL, window = kernel, width,
        #        give.Rkern = FALSE,
        #        n = 512, from, to, cut = 3, na.rm = FALSE, ...)
        # bins <- density(variable)$bw
        # $bw = the bandwidth used.
        #
        # > bw.nrd0
        # function (x)
        # {
        #     if (length(x) < 2L)
        #         stop("need at least 2 data points")
        #     hi <- sd(x)
        #     if (!(lo <- min(hi, IQR(x)/1.34)))
        #         (lo <- hi) || (lo <- abs(x[1L])) || (lo <- 1)
        #     0.9 * lo * length(x)^(-0.2)
        # }
        # > IQR
        # function (x, na.rm = FALSE, type = 7)
        # diff(quantile(as.numeric(x), c(0.25, 0.75), na.rm = na.rm, names = FALSE,
        #     type = type))

        # TODO: Understand R density()$bw function
        bins = 4
        # bins = density(variable)$bw
        tgt = np.arange(target.min(), target.max(), bins)
    else:
        d_dx = (target.max().abs() + target.min().abs()) / bins
        tgt = np.arange(target.min(), target.max(), d_dx)

    # TODO: Check return of function CDF and ['function'] dict
    CDF = NNS_CDF(variable=variable, degree=degree, target=None, type="CDF", plot=False)["Function"]

    # TODO: find function dy.dx()
    PDF = np.maximum(
        dy.dx(unlist(CDF[:, 1]), unlist(CDF[:, 2]), eval_point=tgt, deriv_method="FD")["First"], 0.0
    )
    if plot:
        # TODO: convert ```{R} plot(tgt, PDF, col = 'steelblue', type = 'l', lwd = 3, xlab = "X", ylab = "Density")```
        # plot(tgt, PDF, col = 'steelblue', type = 'l', lwd = 3, xlab = "X", ylab = "Density")
        pass
    # TODO: convert ```{R} return (data.table::data.table(cbind("Intervals" = tgt, PDF))) ```
    return pd.DataFrame()


def NNS_CDF(
    variable: [pd.Series, pd.DataFrame],
    degree: [float, int] = 0,
    target: [float, int, None] = None,
    type: str = "CDF",
    plot: bool = True,
) -> dict:
    # TODO: CONVERT/TEST
    if target is not None:
        # TODO: use iloc[] / loc[]
        if isinstance(variable, pd.Series) or variable.shape[1] == 1:
            if target < variable.min() or target > variable.max():
                raise ValueError(
                    "Please make sure target is within the observed values of variable."
                )
        else:
            if target[1] < variable[:, 1].min() or target[1] > variable[:, 1].max():
                raise ValueError(
                    "Please make sure target 1 is within the observed values of variable 1."
                )
            if target[2] < variable[:, 2].min() or target[2] > variable[:, 2].max():
                raise ValueError(
                    "Please make sure target 2 is within the observed values of variable 2."
                )

    cdf_type = type.lower().strip()
    if cdf_type not in ["cdf", "survival", "hazard", "cumulative hazard"]:
        raise ValueError("Please select a type from: [CDF, survival, hazard, cumulative hazard]")

    if isinstance(variable, pd.Series) or variable.shape[1] == 1:
        overall_target = variable.sort_values()
        x = overall_target
        if degree > 0:
            # TODO: check if target can be an pd.series/np.ndarray
            CDF = LPM_ratio(degree=degree, target=overall_target, variable=variable)
        else:
            # TODO: ecdf function
            cdf_fun = ecdf(x)
            CDF = cdf_fun(overall_target)

        values = pd.DataFrame({variable.name: variable.sort_values(), "CDF": CDF})
        if target is not None:
            P = LPM_ratio(degree=degree, target=target, variable=variable)
        else:
            P = None

        ylabel = "Probability"

        if cdf_type == "survival":
            CDF = 1 - CDF
            P = 1 - P
        elif cdf_type == "hazard":
            # TODO: R density function
            CDF = np.exp(np.log(density(x, n=x.shaṕe[1])["y"]) - np.log(1 - CDF))
            ylabel = "h(x)"
            P = NNS_reg(
                x[-x.shape[1]],
                CDF[-x.shape[1]],
                order="max",
                point_est=c(x[x.shape[1]], target),
                plot=False,
            )["Point.est"]
            CDF[np.isinf(CDF)] = P[1]
            P = P[-1]
        elif cdf_type == "cumulative hazard":
            CDF = -np.log((1 - CDF))
            ylabel = "H(x)"
            P = NNS_reg(
                x[-x.shape[1]],
                CDF[-x.shape[1]],
                order="max",
                point_est=c(x[x.shape[1]], target),
                plot=False,
            )["Point.est"]
            CDF[np.isinf(CDF)] = P[1]
            P = P[-1]
        if plot:
            # TODO matplotlib like:
            pass
            if False:
                plot(
                    x,
                    CDF,
                    pch=19,
                    col="steelblue",
                    xlab=deparse(substitute(variable)),
                    ylab=ylabel,
                    main=toupper(type),
                    type="s",
                    lwd=2,
                )
                points(x, CDF, pch=19, col="steelblue")
                lines(x, CDF, lty=2, col="steelblue")
                if target is not None:
                    segments(target, 0, target, P, col="red", lwd=2, lty=2)
                    segments(min(variable), P, target, P, col="red", lwd=2, lty=2)
                    points(target, P, col="green", pch=19)
                    mtext(text=round(P, 4), col="red", side=2, at=P, las=2)
                    mtext(text=round(target, 4), col="red", side=1, at=target, las=1)
        values = pd.DataFrame({variable.name: x, ylabel: CDF})
        return {"Function": values, "target.value": P}
    else:
        overall_target_1 = variable[:, 1]
        overall_target_2 = variable[:, 2]
        CDF = Co_LPM(
            degree_x=degree,
            degree_y=degree,
            x=variable[:, 1].sort_values(),
            y=variable[:, 2].sort_values(),
            target_x=overall_target_1,
            target_y=overall_target_2,
        ) / (
            Co_LPM(
                degree_x=degree,
                degree_y=degree,
                x=variable[:, 1].sort_values(),
                y=variable[:, 2].sort_values(),
                target_x=overall_target_1,
                target_y=overall_target_2,
            )
            + Co_UPM(
                degree_x=degree,
                degree_y=degree,
                x=variable[:, 1].sort_values(),
                y=variable[:, 2].sort_values(),
                target_x=overall_target_1,
                target_y=overall_target_2,
            )
            + D_UPM(
                degree_x=degree,
                degree_y=degree,
                x=variable[:, 1].sort_values(),
                y=variable[:, 2].sort_values(),
                target_x=overall_target_1,
                target_y=overall_target_2,
            )
            + D_LPM(
                degree_x=degree,
                degree_y=degree,
                x=variable[:, 1].sort_values(),
                y=variable[:, 2].sort_values(),
                target_x=overall_target_1,
                target_y=overall_target_2,
            )
        )
        if cdf_type == "survival":
            CDF = 1 - CDF
        elif cdf_type == "hazard":
            CDF = variable.sort_values() / (1 - CDF)
        elif cdf_type == "cumulative hazard":
            CDF = -np.log(1 - CDF)
        if target is not None:
            P = Co_LPM(
                degree_x=degree,
                degree_y=degree,
                x=variable[:, 1],
                y=variable[:, 2],
                target_x=target[1],
                target_y=target[2],
            ) / (
                Co_LPM(
                    degree_x=degree,
                    degree_y=degree,
                    x=variable[:, 1],
                    y=variable[:, 2],
                    target_x=target[1],
                    target_y=target[2],
                )
                + Co_UPM(
                    degree_x=degree,
                    degree_y=degree,
                    x=variable[:, 1],
                    y=variable[:, 2],
                    target_x=target[1],
                    target_y=target[2],
                )
                + D_LPM(
                    degree_x=degree,
                    degree_y=degree,
                    x=variable[:, 1],
                    y=variable[:, 2],
                    target_x=target[1],
                    target_y=target[2],
                )
                + D_UPM(
                    degree_x=degree,
                    degree_y=degree,
                    x=variable[:, 1],
                    y=variable[:, 2],
                    target_x=target[1],
                    target_y=target[2],
                )
            )
        else:
            P = None

        if plot:
            # TODO: Matplotlib like
            pass
            if False:
                plot3d(
                    variable[:, 1],
                    variable[:, 2],
                    CDF,
                    col="steelblue",
                    xlab=variable[:, 1].name,
                    ylab=variable[:, 2].name,
                    zlab="Probability",
                    box=False,
                    pch=19,
                )
                if target is not None:
                    points3d(target[1], target[2], P, col="green", pch=19)
                    points3d(target[1], target[2], 0, col="red", pch=15, cex=2)
                    lines3d(
                        x=c(target[1], variable[:, 1].max()),
                        y=c(target[2], variable[:, 2].max()),
                        z=c(P, P),
                        col="red",
                        lwd=2,
                        lty=3,
                    )
                    lines3d(
                        x=c(target[1], target[1]),
                        y=c(target[2], target[2]),
                        z=c(0, P),
                        col="red",
                        lwd=1,
                        lty=3,
                    )
                    text3d(
                        variable[:, 1].max(),
                        variable[:, 2].max(),
                        P,
                        texts=f"P = {round(P, 4)}",
                        pos=4,
                        col="red",
                    )
    return {"CDF": pd.DataFrame({"variable": variable, "CDF": CDF}), "P": P}


__all__ = [
    "pd_fill_diagonal",
    "LPM",
    "UPM",
    "Co_UPM",
    "Co_LPM",
    "D_LPM",
    "D_UPM",
    "PM_matrix",
    "LPM_ratio",
    "UPM_ratio",
    # "NNS_PDF", # TODO
    # "NNS_CDF", # TODO
]



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
from .Partial_Moments import LPM_ratio
import scipy.optimize


def _LPM_VaR(
    percentile: [float, int], degree: [float, int, str, None], x: [pd.Series, np.ndarray, list]
) -> float:
    x_min = np.min(x)
    x_max = np.max(x)
    if x_min == x_max:
        return x_min

    percentile = max(min(percentile, 1.0), 0.0)
    # if percentile <= 0:
    #    return x_min
    # elif percentile >= 1:
    #    return x_max
    # force float variable
    if degree == 0:
        # td = tdigest.TDigest()
        # td.batch_update(x)
        # return td.percentile(percentile)
        return np.quantile(x, percentile, interpolation="linear")
    # degree > 0
    x = np.array(x, dtype=float)
    x_max = float(x_max)
    x_min = float(x_min)
    x0 = np.mean(x)
    x_range = x_max - x_min
    x1 = x_min if x_range == 0 else (x_range * percentile + x_min)

    def _func(b: float) -> float:
        return LPM_ratio(degree, max(x_min, min(x_max, b)), x) - percentile

    ret = scipy.optimize.root_scalar(
        f=_func,
        # method='bisect',
        bracket=[x_min, x_max],
        x0=x0,
        x1=x1,
    )
    if not ret.converged:
        raise Exception(f"Root find didn't converged: {ret}")
    return ret.root


_vec_LPM_VaR = np.vectorize(_LPM_VaR, excluded=["degree", "x"])


def LPM_VaR(
    percentile: [float, int, np.array, pd.Series, list],
    degree: [float, int, str, None],
    x: [pd.Series, np.ndarray, list],
) -> [float, np.array]:
    r"""
    LPM VaR

    Generates a value at risk (VaR) quantile based on the Lower Partial Moment ratio.

    @param percentile numeric [0, 1]; The percentile for left-tail VaR (vectorized).
    @param degree integer; \code{(degree = 0)} for discrete distributions, \code{(degree = 1)} for continuous distributions.
    @param x a numeric vector.
    @return Returns a numeric value representing the point at which \code{"percentile"} of the area of \code{x} is below.
    @author Fred Viole, OVVO Financial Systems
    @references Viole, F. and Nawrocki, D. (2013) "Nonlinear Nonparametric Statistics: Using Partial Moments"
    \url{https://www.amazon.com/dp/1490523995}
    @examples
    set.seed(123)
    x <- rnorm(100)

    ## For 5% quantile, left-tail
    LPM.VaR(0.05, 0, x)
    @export
    """
    func = _LPM_VaR
    if isinstance(percentile, (np.ndarray, pd.Series, list)):
        func = _vec_LPM_VaR
    return func(percentile=percentile, degree=degree, x=x)


def _UPM_VaR(
    percentile: [float, int],
    degree: [float, int, str, None],
    x: [pd.Series, np.array],
) -> float:
    x_min = np.min(x)
    x_max = np.max(x)
    if x_min == x_max:
        return x_min

    percentile = max(min(percentile, 1.0), 0.0)
    # if percentile <= 0:
    #    return x_max
    # elif percentile >= 1:
    #    return x_min
    # force float variable
    if degree == 0:
        # td = tdigest.TDigest()
        # td.batch_update(x)
        # return td.percentile(percentile)
        return np.quantile(x, 1 - percentile, interpolation="linear")

    # degree > 0
    x = np.array(x, dtype=float)
    x_max = float(x_max)
    x_min = float(x_min)
    x0 = np.mean(x)
    x_range = x_max - x_min
    x1 = x_min if x_range == 0 else (x_range * (1 - percentile) + x_min)

    def _func(b: float) -> float:
        return LPM_ratio(degree, max(x_min, min(x_max, b)), x) - (1 - percentile)

    ret = scipy.optimize.root_scalar(
        f=_func,
        # method='bisect',
        bracket=[x_min, x_max],
        x0=x0,
        x1=x1,
    )
    if not ret.converged:
        raise Exception(f"Root find didn't converged: {ret}")
    return ret.root


_vec_UPM_VaR = np.vectorize(_UPM_VaR, excluded=["degree", "x"])


def UPM_VaR(
    percentile: [float, int, np.array, pd.Series, list],
    degree: [float, int, str, None],
    x: [pd.Series, np.array],
) -> [float, np.array]:
    r"""
    UPM VaR

    Generates an upside value at risk (VaR) quantile based on the Upper Partial Moment ratio
    @param percentile numeric [0, 1]; The percentile for right-tail VaR (vectorized).
    @param degree integer; \code{(degree = 0)} for discrete distributions, \code{(degree = 1)} for continuous distributions.
    @param x a numeric vector.
    @return Returns a numeric value representing the point at which \code{"percentile"} of the area of \code{x} is above.
    @examples
    set.seed(123)
    x <- rnorm(100)

    ## For 5% quantile, right-tail
    UPM.VaR(0.05, 0, x)
    @export
    """
    func = _UPM_VaR
    if isinstance(percentile, (np.ndarray, pd.Series, list)):
        func = _vec_UPM_VaR
    return func(percentile=percentile, degree=degree, x=x)


__all__ = [
    "LPM_VaR",
    "UPM_VaR",
]



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
from .Uni_SD_Routines import NNS_FSD_uni, NNS_SSD_uni, NNS_TSD_uni
from .Partial_Moments import LPM


def NNS_SD_efficient_set(
    x: [pd.DataFrame, np.ndarray],
    degree: int,
    type_first_degree: str = "discrete",
    status: bool = True,
) -> [list, np.ndarray]:
    type_first_degree = type_first_degree.lower()

    if len(x.shape) != 2:
        raise Exception(f"x shape should contains 2 elements (dataframe like): {x.shape}")

    if type_first_degree not in ["discrete", "continuous"]:
        raise Exception("type_first_degree needs to be either 'discrete' or 'continuous'")

    if degree not in [1, 2, 3]:
        raise Exception("degree needs to be 1, 2, or 3")

    n = x.shape[1]
    max_target = np.max(np.max(x))

    Dominated_set = []
    current_base = []
    if isinstance(x, pd.DataFrame):
        LPM_order = [LPM(1, max_target, x.values[:, i]) for i in range(n)]
        LPM_order_argsort = np.argsort(LPM_order)
        final_ranked = x.values[:, LPM_order_argsort]
    else:
        LPM_order = [LPM(1, max_target, x[:, i]) for i in range(n)]
        LPM_order_argsort = np.argsort(LPM_order)
        final_ranked = x[:, LPM_order_argsort]
    current_base.append(0)
    for i in range(n - 1):
        if status:
            print(f"Checking {i+1} of {n-1}\r", end="")
            if i == (n - 1):
                print("                                        ", end="\n")
        base, challenger = final_ranked[:, current_base[-1]], final_ranked[:, i + 1]
        sd_test = 0
        if degree == 1:
            sd_test = NNS_FSD_uni(base, challenger, type_test=type_first_degree)
        elif degree == 2:
            sd_test = NNS_SSD_uni(base, challenger)
        elif degree == 3:
            sd_test = NNS_TSD_uni(base, challenger)
        if sd_test == 1:
            Dominated_set.append(i + 1)
        elif sd_test == 0:
            sd_found = False
            for j in current_base:
                base = final_ranked[:, j]
                new_base_sd_test = 0
                if degree == 1:
                    new_base_sd_test = NNS_FSD_uni(base, challenger, type_test=type_first_degree)
                elif degree == 2:
                    new_base_sd_test = NNS_SSD_uni(base, challenger)
                elif degree == 3:
                    new_base_sd_test = NNS_TSD_uni(base, challenger)
                if new_base_sd_test != 0:
                    sd_found = True
                    Dominated_set.append(i + 1)
                    break
            if not sd_found:
                current_base.append(i + 1)

    if len(Dominated_set) > 0:
        if isinstance(x, pd.DataFrame):
            return list(
                x.columns[
                    LPM_order_argsort[
                        [i for i in range(len(LPM_order_argsort)) if i not in Dominated_set]
                    ]
                ]
            )
        else:
            return LPM_order_argsort[
                [i for i in range(len(LPM_order_argsort)) if i not in Dominated_set]
            ]
    elif isinstance(x, pd.DataFrame):
        return list(x.columns[LPM_order_argsort])
    else:
        return LPM_order_argsort


__all__ = [
    "NNS_SD_efficient_set",
]



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
import scipy
import scipy.stats
import KDEpy


def fivenum(v: [pd.Series, np.ndarray]) -> list:
    """Tukey Five-Number Summaries

    Returns Tukey's five number summary (minimum, lower-hinge, median, upper-hinge, maximum) for the input data."""
    x = v[~np.isnan(v)]
    x = np.sort(x)
    n = x.shape[0]
    if n == 0:
        return [np.nan, np.nan, np.nan, np.nan, np.nan]

    n4 = np.floor((n + 3) / 2) / 2
    d = np.array([1, n4, (n + 1) / 2, n + 1 - n4, n])
    return list(0.5 * (x[np.floor(d).astype(int) - 1] + x[np.ceil(d).astype(int) - 1]))


def bw_nrd0(x: [scipy.stats.kde.gaussian_kde, np.ndarray, pd.Series]):
    """Kernel density bandwidth estimator - like R default option of density function"""
    # https://stackoverflow.com/questions/38337804/is-there-a-scipy-numpy-alternative-to-rs-nrd0
    if isinstance(x, pd.Series):
        tmpx = x.values if len(x.shape) == 1 else x.values[0, :]
    elif isinstance(x, np.ndarray):
        tmpx = x if len(x.shape) == 1 else x[0, :]
    else:
        tmpx = x.dataset if len(x.dataset.shape) == 1 else x.dataset[0, :]
    if len(tmpx) < 2:
        raise (Exception("need at least 2 data points"))
    hi = np.std(tmpx, ddof=1)
    q75, q25 = np.percentile(tmpx, [75, 25])
    iqr = q75 - q25
    lo = min(hi, iqr / 1.34)
    if not (lo != 0):
        if hi != 0:
            lo = hi
        elif abs(x[0]) != 0:
            lo = abs(x[0])
        else:
            lo = 1
    bw = 0.9 * lo * len(tmpx) ** -0.2
    return bw


def mode(x: [pd.Series, np.ndarray]) -> float:
    """Continuous Mode of a distribution"""

    #    d = tryCatch(
    #        density(
    #            as.numeric(x),
    #            na.rm = TRUE,
    #            n = 128,
    #            from = min(x),
    #            to = max(x)),
    #            error = function(e) (
    #                median(x) + mean(fivenum(x)[2:4])
    #            )/2
    #    )
    #    tryCatch(
    #        d$x[which.max(d$y)],
    #        error = function(e) d
    #    )

    if x.shape[0] == 2:
        return np.median(x)

    if (
        True
    ):  # TODO Check if KDEpy exists, if not use the old scipy.stats.gaussian_kde (with bad outputs btw)
        if isinstance(x, pd.Series):
            a, b = KDEpy.FFTKDE(kernel="gaussian", bw=bw_nrd0(x.values)).fit(x.values).evaluate()
        else:
            a, b = KDEpy.FFTKDE(kernel="gaussian", bw=bw_nrd0(x)).fit(x).evaluate()
        mode_value = a[np.argmax(b)]
    elif (
        False
    ):  # TODO Check if KDEpy exists, if not use the old scipy.stats.gaussian_kde (with bad outputs btw)
        # https://rmflight.github.io/post/finding-modes-using-kernel-density-estimates/
        kernel = scipy.stats.gaussian_kde(x, bw_method=bw_nrd0)
        height = kernel.pdf(x)
        mode_value = x[np.argmax(height)]
    return mode_value


def mode_class(x: [np.ndarray, pd.Series]) -> float:
    """Classification Mode of a distribution"""
    if x.shape[0] == 0:
        return np.nan
    unique, indexes, counts = np.unique(x[~np.isnan(x)], return_counts=True, return_index=True)
    max_uniques = unique[counts == np.max(counts)]
    return x[np.isin(x, max_uniques)][0]


def gravity(x: [pd.Series, np.ndarray]) -> float:
    """Central Tendency"""
    return (np.mean(x) + np.median(x) + mode(x)) / 3.0


def gravity_class(x: [pd.Series, np.ndarray]) -> float:
    """Central Tendency Class"""
    return (np.mean(x) + np.mean(fivenum(x)[1:4])) / 2.0


def alt_cbind(
    x: [pd.Series, np.ndarray], y: [pd.Series, np.ndarray], first: bool = False
) -> pd.DataFrame:
    """### cbind different length vectors"""
    if not isinstance(x, pd.Series):
        x = pd.Series(x)
    if not isinstance(y, pd.Series):
        y = pd.Series(y)
    len_x = len(x)
    len_y = len(y)
    if len_x < len_y:
        v = pd.Series(np.full(fill_value=np.nan, shape=(len_y - len_x), dtype=float))
        if first:
            x = pd.Series(v).append(x).rename(x.name)
        else:
            x = pd.concat([x, v], axis=0).rename(x.name)
        i = len_x
        while i < len_y:
            next_i = i + len_x
            if next_i >= len_y:
                next_i = len_y
            x[i:next_i] = x[0 : next_i - i]
            if next_i >= len_y:
                break
            i = next_i

    if len_y < len_x:
        v = pd.Series(np.full(fill_value=np.nan, shape=(len_x - len_y), dtype=float))
        if first:
            y = pd.Series(v).append(y).rename(y.name)
        else:
            y = pd.concat([y, v], axis=0).rename(y.name)
        i = len_y
        while i < len_x:
            next_i = i + len_y
            if next_i >= len_x:
                next_i = len_x
            y[i:next_i] = y[0 : next_i - i]
            if next_i >= len_x:
                break
            i = next_i
    df = pd.DataFrame({0: x.values, 1: y.values})
    df.columns = (0 if x.name is None else x.name, 1 if y.name is None else y.name)
    return df


def factor_2_dummy(x: pd.Series) -> [pd.Series, pd.DataFrame]:
    """Factor to dummy variable"""
    if x.dtype.name == "category" and x.nunique() > 1:
        return pd.get_dummies(x, prefix=x.name).iloc[:, 1:]
    return x.astype(float)


def factor_2_dummy_FR(x: pd.Series) -> [pd.Series, pd.DataFrame]:
    """Factor to dummy variable FULL RANK"""
    if x.dtype.name == "category" and x.nunique() > 1:
        return pd.get_dummies(x, prefix=x.name)
    return x.astype(float)


def generate_vectors(
    x: [np.ndarray, pd.Series], l: [np.ndarray, pd.Series, list, float, int]
) -> dict:
    r"""Generator for 1:length(lag) vectors in NNS.ARMA"""
    if isinstance(l, (int, float)):
        l = [l]
    Component_series = {}
    Component_index = {}
    len_x = len(x)
    len_l = len(l)
    for i in range(len_l):
        CS = np.array([x[i] for i in range(len_x, -1, -l[i]) if len_x > i >= 0])[::-1]
        CS = CS[~np.isnan(CS)]
        Component_series[f"Series.{i+1}"] = CS
        Component_index[f"Index.{i+1}"] = np.arange(1, len(CS) + 1)
    return {"Component.index": Component_index, "Component.series": Component_series}


def ARMA_seas_weighting(sf: [int, float, bool], mat: dict) -> dict:
    # TODO
    ### Weight and lag function for seasonality in NNS.ARMA
    sf_true = sf is True or (not isinstance(sf, bool) and sf != 0)
    if (
        "all.periods" not in mat
        or "best.period" not in mat
        or "periods" not in mat
        or len(mat["all.periods"].shape) != 2
    ):
        raise Exception("An matrix was expected at mat['all.periods']")

    n = mat["all.periods"].shape[0]
    if n <= 0:
        return {"lag": mat["all.periods"]["Period"], "Weights": 1}
    elif n == 1:
        return {"lag": 1, "Weights": 1}
    else:
        if sf_true:
            return {"lag": mat["all.periods"]["Period"].iloc[0], "Weights": 1}
        # Determine lag from seasonality test

        lag = mat["all.periods"]["Period"][~np.isnan(mat["all.periods"]["Period"])]
        Observation_weighting = 1 / np.sqrt(lag)
        if (
            np.any(np.isnan(mat["all.periods"]["Coefficient.of.Variation"]))
            and len(mat["all.periods"]["Coefficient.of.Variation"]) == 1
        ):
            Lag_weighting = 1
        else:
            Lag_weighting = (
                mat["all.periods"]["Variable.Coefficient.of.Variation"]
                - mat["all.periods"]["Coefficient.of.Variation"]
            )
        Weights = (Lag_weighting * Observation_weighting) / np.sum(
            Lag_weighting * Observation_weighting
        )
        return {"lag": lag, "Weights": Weights}


def lag_mtx(x, tau):
    # TODO: Translate
    r"""
    Lag matrix generator for NNS.VAR
    Vector of tau for single different tau per variables tau = c(1, 4)
    List of tau vectors for multiple different tau per variables tau = list(c(1,2,3), c(4,5,6))
    """
    colheads = None
    # TODO: Translate ```{R} max(unlist(tau)) ```
    max_tau = max(unlist(tau))
    if is_null(dim(x)[2]):
        # TODO: Translate ```{R} noquote(as_character(deparse(substitute(x)))) ```
        colheads = noquote(as_character(deparse(substitute(x))))
        x = t(t(x))
    j_vectors = []
    # TODO: Convert ```{R} ncol(x)```
    for j in range(ncol(x)):
        if is_null(colheads):
            # TODO: Convert
            colheads = colnames(x)[j]
            colheads = noquote(as_character(deparse(substitute(colheads))))

        x.vectors = list()
        heads = paste0(colheads, "_tau_")
        heads = gsub('"', "", heads)

        for i in range(max_tau):
            x_vectors[[paste(heads, i, sep="")]] = numeric(0)
            start = max_tau - i + 1
            end = length(x[:, j]) - i
            x_vectors[[i + 1]] = x[start:end, j]

        j_vectors[[j]] = do.call(cbind, x.vectors)
        colheads = None
    mtx = pd.DataFrame(do.call(cbind, j.vectors))

    relevant_lags = list(length(tau))
    if length(unlist(tau)) > 1:
        for i in range(length(tau)):
            relevant_lags[[i]] = c((i - 1) * max_tau + i, (i - 1) * max_tau + unlist(tau[[i]]) + i)
        relevant_lags = sort(unlist(relevant_lags))
        mtx = mtx[:, relevant_lags]
    vars = which(grepl("tau_0", colnames(mtx)))

    everything_else = seq_len(dim(mtx)[2])[-vars]
    mtx = mtx[:, c(vars, everything_else)]
    return mtx


def RP(x: pd.Series, rows=None, cols=None, na_rm=False):
    r"""Row products for NNS.dep.hd"""
    if rows is not None and cols is not None:
        pass
        # x = x[rows, cols, drop = False]
    elif rows is not None:
        pass
        # x = x[rows, , drop = False]
    elif cols is not None:
        pass
        # x = x[, cols, drop = False]

    n = nrow(x)
    y = double(length=n)
    if n == 0:
        return y
    for ii in range(seq_len(n)):
        pass
        # y[ii] = prod(x[ii, , drop = True], na_rm = na_rm)
    return y


def NNS_meboot_part(x: pd.Series, n, z, xmin, xmax, desintxb, reachbnd):
    r"""Refactored meboot::meboot.part function using tdigest"""
    # Generate random numbers from the [0,1] uniform interval
    # TODO: Convert
    p = runif(n, min=0, max=1)
    # Assign quantiles of x from p
    # td = tdigest::tdigest(x, compression = max(100, np.log(n,10)*100))
    q = np.quantile(x, p)
    # try:
    #    q = tdigest::tquantile(td, p)
    # except Exception:
    #    q = x.quantile(p)
    # TODO: Convert
    ref1 = which(p <= (1 / n))
    if length(ref1) > 0:
        qq = approx(c(0, 1 / n), c(xmin, z[1]), p[ref1])["y"]
        q[ref1] = qq
        if not reachbnd:
            q[ref1] = qq + desintxb[1] - 0.5 * (z[1] + xmin)
    ref4 = which(p == ((n - 1) / n))
    if len(ref4) > 0:
        q[ref4] < -z[n - 1]

    ref5 = which(p > ((n - 1) / n))
    if len(ref5) > 0:
        # Right tail proportion p[i]
        qq = approx(c((n - 1) / n, 1), c(z[n - 1], xmax), p[ref5])["y"]
        q[ref5] = qq  # this implicitly shifts xmax for algorithm
        if not reachbnd:
            q[ref5] = qq + desintxb[n] - 0.5 * (z[n - 1] + xmax)
        # such that the algorithm gives xmax when p[i]=1
        # this is the meaning of reaching the bounds xmax and xmin
    return q


def NNS_meboot_expand_sd(x: pd.Series, ensemble, fiv=5):
    # TODO: Convert
    r"""Refactored meboot::expand.sd function"""
    if is_null(ncol(x)):
        sdx = x.std()
    else:
        sdx = apply(x, 2, np.std)

    sdf = c(sdx, apply(ensemble, 2, np.std))
    sdfa = sdf / sdf[1]  # ratio of actual sd to that of original data
    sdfd = sdf[1] / sdf  # ratio of desired sd to actual sd

    # expansion is needed since some of these are <1 due to attenuation
    mx = 1 + (fiv / 100)
    # following are expansion factors
    id = which(sdfa < 1)
    if len(id) > 0:
        sdfa[id] = runif(n=length(id), min=1, max=mx)

    sdfdXsdfa = sdfd[-1] * sdfa[-1]  #  TODO: this looks like some keyboard with problem ahhaha
    id = which(np.floor(sdfdXsdfa) > 0)
    if len(id) > 0:
        if len(id) > 1:
            pass
            # ensemble[:,id] = ensemble[:,id].dot(np.diag(sdfdXsdfa[id]))    # TODO convert ```{R} diag(sdfdXsdfa[id])```
        else:
            pass
            # ensemble[:,id] = ensemble[:,id] * sdfdXsdfa[id]
    if is_ts(x):
        ensemble = ts(ensemble, frequency=frequency(x), start=start(x))
    return ensemble


__all__ = [
    "bw_nrd0",
    "mode",
    "mode_class",
    "gravity",
    "alt_cbind",
    "factor_2_dummy",
    "factor_2_dummy_FR",
    "generate_vectors",
    "ARMA_seas_weighting",
    "lag_mtx",
    "RP",
    "NNS_meboot_part",
    "NNS_meboot_expand_sd",
]



# -*- coding: utf-8 -*-
from .Binary_ANOVA import *
from .Copula import *

# from .dy_dx import *
from .FSD import *

from . import Internal_Functions
from .LPM_UPM_VaR import *
from .NNS_term_matrix import *
from .Numerical_Differentiation import *
from .Partial_Moments import *
from .SD_Efficient_Set import *
from .SSD import *
from .TSD import *
from .Uni_SD_Routines import *

__version__ = "0.1.10"



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd

def NNS_part(
    x: [np.ndarray, pd.Series], y: [np.ndarray, pd.Series],
    Voronoi: bool=False,
    _type=None,
    order=None,
    obs_req: int = 8,
    min_obs_stop: bool = True,
    noise_reduction: str = "off"
):
    """
    NNS Partition Map

    Creates partitions based on partial moment quadrant centroids, iteratively assigning identifications to observations based on those quadrants (unsupervised partitional and hierarchial clustering method).  Basis for correlation, dependence \link{NNS.dep}, regression \link{NNS.reg} routines.

    @param x a numeric vector.
    @param y a numeric vector with compatible dimensions to \code{x}.
    @param Voronoi logical; \code{FALSE} (default) Displays a Voronoi type diagram using partial moment quadrants.
    @param _type \code{NULL} (default) Controls the partitioning basis.  Set to \code{(type = "XONLY")} for X-axis based partitioning.  Defaults to \code{NULL} for both X and Y-axis partitioning.
    @param order integer; Number of partial moment quadrants to be generated.  \code{(order = "max")} will institute a perfect fit.
    @param obs_req integer; (8 default) Required observations per cluster where quadrants will not be further partitioned if observations are not greater than the entered value.  Reduces minimum number of necessary observations in a quadrant to 1 when \code{(obs.req = 1)}.
    @param min_obs_stop logical; \code{TRUE} (default) Stopping condition where quadrants will not be further partitioned if a single cluster contains less than the entered value of \code{obs.req}.
    @param noise_reduction the method of determining regression points options for the dependent variable \code{y}: ("mean", "median", "mode", "off"); \code{(noise.reduction = "mean")} uses means for partitions.  \code{(noise.reduction = "median")} uses medians instead of means for partitions, while \code{(noise.reduction = "mode")} uses modes instead of means for partitions.  Defaults to \code{(noise.reduction = "off")} where an overall central tendency measure is used, which is the default for the independent variable \code{x}.
    @return Returns:
     \itemize{
      \item{\code{"dt"}} a \link{data.table} of \code{x} and \code{y} observations with their partition assignment \code{"quadrant"} in the 3rd column and their prior partition assignment \code{"prior.quadrant"} in the 4th column.
      \item{\code{"regression.points"}} the \link{data.table} of regression points for that given \code{(order = ...)}.
      \item{\code{"order"}}  the \code{order} of the final partition given \code{"min.obs.stop"} stopping condition.
     }

    @note \code{min.obs.stop = FALSE} will not generate regression points due to unequal partitioning of quadrants from individual cluster observations.

    @author Fred Viole, OVVO Financial Systems
    @references Viole, F. and Nawrocki, D. (2013) "Nonlinear Nonparametric Statistics: Using Partial Moments"
    \url{https://www.amazon.com/dp/1490523995/ref=cm_sw_su_dp}
    @examples
    set.seed(123)
    x <- rnorm(100) ; y <- rnorm(100)
    NNS.part(x, y)

    ## Data.table of observations and partitions
    NNS.part(x, y, order = 1)$dt

    ## Regression points
    NNS.part(x, y, order = 1)$regression.points

    ## Voronoi style plot
    NNS.part(x, y, Voronoi = TRUE)

    ## Examine final counts by quadrant
    DT <- NNS.part(x, y)$dt
    DT[ , counts := .N, by = quadrant]
    DT
    @export
    """


    noise_reduction = noise_reduction.lower()
    if noise_reduction not in ["mean", "median", "mode", "off", "mode_class"]:
        raise Exception("Please ensure noise_reduction is from  [mean, median, mode, off, mode_class]")

    if isinstance(x, pd.Series):
        x = x.values
    if isinstance(y, pd.Series):
        y = y.values

    if obs_req is None:
        obs_req = 8

    if not order is None and order == 0:
        order = 1
    if Voronoi:
        # TODO
        x.label = deparse(substitute(x))
        y.label = deparse(substitute(y))

    # TODO
    PART = data.table::data.table(
        x, y,
        quadrant = "q",
        prior.quadrant = "pq"
    )[, `:=`(counts, .N), by = "quadrant"][, `:=`(old.counts, .N), by = "prior.quadrant"]

    if Voronoi:
        # TODO: matplotlib
        plot(x, y, col = "steelblue", cex.lab = 1.5, xlab = x.label, ylab = y.label)

    if x.shape[0] <= 8:
        if order is None:
            order = 1
            hard_stop = max(np.ceil(np.log(x.shape[0], 2)), 1)
        else:
            obs_req = 0
            hard_stop = x.shape[0]

    if order is None:
        order = np.max(np.ceil(np.log(x.shape[0], 2)), 1)

    # TODO is numeric?
    #if(!is.numeric(order)):
    if False:
        obs_req = 0
        hard_stop = np.max(np.ceil(np.log(x.shape[0], 2)), 1) + 2
    else:
        #obs_req = obs_req
        hard_stop = 2*np.max(np.ceil(np.log(x.shape[0], 2)), 1) + 2

    RP = None
    if _type is None:
        i = 0
        while (i >= 0):
            if i == order or i == hard_stop:
                break
            # TODO
            PART[counts >= obs_req, `:=`(counts, .N), by = quadrant]
            PART[old_counts >= obs_req, `:=`(old_counts, .N), by = prior_quadrant]
            l_PART = max(PART$counts)

            obs_req_rows = PART[counts >= obs_req, which = TRUE]
            old_obs_req_rows = PART[old_counts >= obs_req, which = TRUE]

            if len(obs_req_rows)==0:
                break
            if min_obs_stop and obs_req > 0 and (len(obs_req_rows) < len(old_obs_req_rows)):
                break

            if noise_reduction == "off":
                if Voronoi:
                    if l_PART > obs_req:
                        PART[obs_req_rows, {
                                segments(min(x), gravity(y), max(x), gravity(y), lty = 3)
                                segments(gravity(x), min(y), gravity(x), max(y), lty = 3)
                        }, by = quadrant]
                 RP = PART[
                    obs.req.rows,
                    mapply(function(f,z) as.list(f(z)), list(x = gravity, y = gravity), .SD),
                    by = quadrant,
                    .SDcols = x:y
                ]

            if noise.reduction == "mean":
                if Voronoi:
                    if l_PART > obs_req:
                        PART[
                            obs.req.rows, {
                            segments(min(x), mean(y), max(x), mean(y), lty = 3)
                            segments(gravity(x), min(y), gravity(x), max(y), lty = 3)
                        }, by = quadrant]
                RP = PART[
                    obs.req.rows,
                    mapply(function(f,z) as.list(f(z)), list(x = gravity, y = mean), .SD),
                    by = quadrant,
                    .SDcols = x:y
                ]

            if noise_reduction == "median":
                if Voronoi:
                    if l_PART > obs_req:
                        PART[
                            obs.req.rows, {
                                segments(min(x), median(y), max(x), median(y), lty = 3)
                                segments(gravity(x), min(y), gravity(x), max(y), lty = 3)
                            },
                            by = quadrant
                        ]
                RP = PART[
                    obs.req.rows,
                    mapply(function(f,z) as.list(f(z)), list(x = gravity, y = median), .SD),
                    by = quadrant,
                    .SDcols = x:y
                ]

            if noise_reduction == "mode":
                if Voronoi:
                    if l_PART > obs_req:
                        PART[
                            obs.req.rows, {
                                segments(min(x), mode(y), max(x), mode(y), lty = 3)
                                segments(gravity(x), min(y), gravity(x), max(y), lty = 3)
                            },
                            by = quadrant
                        ]
                RP = PART[
                    obs.req.rows,
                    mapply(function(f,z) as.list(f(z)), list(x = gravity, y = mode), .SD),
                    by = quadrant,
                    .SDcols = x:y
                ]

            if noise_reduction == "mode_class":
                if Voronoi:
                    if l_PART > obs_req:
                        PART[
                            obs.req.rows, {
                                segments(min(x), mode_class(y), max(x), mode_class(y), lty = 3)
                                segments(gravity_class(x), min(y), gravity_class(x), max(y), lty = 3)
                            },
                            by = quadrant
                        ]
                RP = PART[
                    obs.req.rows,
                    mapply(function(f,z) as.list(f(z)), list(x = gravity_class, y = mode_class), .SD),
                    by = quadrant,
                    .SDcols = x:y
                ]

            RP[, `:=`(prior.quadrant, (quadrant))]
            PART[obs.req.rows, `:=`(prior.quadrant, (quadrant))]
            old_parts = length(unique(PART$quadrant))

            PART[
                RP,
                on = .(quadrant), `:=`(
                    q_new, {
                        lox = x.x <= i.x
                        loy = x.y <= i.y
                        1L + lox + loy * 2L
                    }
               )
            ]
            PART[
                obs.req.rows,
                `:=`(quadrant, paste0(quadrant, q_new))
            ]
            new_parts = len(np.unique(PART$quadrant))
            if min(PART$counts) <= obs_req) and i >= 1:
                break
            i += 1

        if RP is None:
            RP = PART[, c("quadrant", "x", "y")]
        # TODO: is.numeric(order) ?
        if (!is.numeric(order) || is.null(dim(RP))):
            RP = PART[, c("quadrant", "x", "y")]
        else:
            RP[, `:=`(prior.quadrant = NULL)]

        PART[, `:=`(counts = NULL, old.counts = NULL, q_new = NULL)]
        RP = data.table::setorder(RP[], quadrant)[]
        if Voronoi:
            # TODO: Matplotlib
            title(main = paste0("NNS Order = ", i), cex.main = 2)
            if min_obs_stop:
                points(RP$x, RP$y, pch = 15, lwd = 2, col = "red")
        if min_obs_stop == False:
            RP = None
        return{
            "order": i,
            "dt": PART[],
            "regression.points": RP
        }

    if not _type is None:
        i = 0
        while i >= 0:
            if i == order or i == hard_stop:
                break
            PART[counts > obs.req/2, `:=`(counts, .N), by = quadrant]
            PART[old.counts > obs.req/2, `:=`(old.counts, .N), by = prior.quadrant]
            obs_req_rows = PART[counts > obs.req/2, which = TRUE]
            old_obs_req_rows = PART[old.counts > obs.req/2, which = TRUE]

            if len(obs_req_rows) == 0:
                break
            if obs_req > 0 and (len(obs_req_rows) < len(old_obs_req_rows)):
                break
            if noise_reduction == "off":
                RP = PART[
                    obs.req.rows,
                    mapply(function(f,z) as.list(f(z)), list(x = gravity, y = gravity), .SD),
                    by = quadrant,
                   .SDcols = x:y
                ]

            if noise_reduction == "mean":
                RP = PART[
                    obs.req.rows,
                    mapply(function(f,z) as.list(f(z)), list(x = gravity, y = mean), .SD),
                    by = quadrant,
                    .SDcols = x:y
                ]

            if noise_reduction == "mode":
                RP = PART[
                    obs.req.rows,
                    mapply(function(f,z) as.list(f(z)), list(x = gravity, y = mode), .SD),
                    by = quadrant,
                    .SDcols = x:y
                ]

            if noise_reduction == "mode_class":
                RP = PART[
                    obs.req.rows,
                    mapply(function(f,z) as.list(f(z)), list(x = gravity_class, y = mode_class), .SD),
                    by = quadrant,
                    .SDcols=x:y
                ]

            if noise_reduction == "median":
                RP = PART[
                    obs.req.rows,
                    mapply(function(f,z) as.list(f(z)), list(x = gravity, y = median), .SD),
                    by = quadrant,
                    .SDcols = x:y
                ]

            RP[, `:=`(prior.quadrant, (quadrant))]
            PART[obs.req.rows, `:=`(prior.quadrant, (quadrant))]
            old_parts = len(np.unique(PART$quadrant))
            PART[
                RP,
                on = .(quadrant), `:=`(q_new, {
                    lox = x.x > i.x
                    1L + lox
                })
            ]
            PART[
                obs_req_rows,
                `:=`(quadrant, paste0(quadrant, q_new))
            ]
            new_parts = len(np.unique(PART$quadrant))
            if min(PART$counts) <= obs_req and i >= 1:
                break
            i += 1

        if RP is None:
            RP = PART[, c("quadrant", "x", "y")]
        # TODO is.numeric(order)?
        if not is_numeric(order) or is_null(dim(RP)):
            RP = PART[, c("quadrant", "x", "y")]
        else:
            RP[, `:=`(prior.quadrant = NULL)]

        PART[, `:=`(counts = NULL, old.counts = NULL, q_new = NULL)]
        RP = data.table::setorder(RP[], quadrant)[]
        if np.mean([
                len(np.unique(np.diff(x))),
                len(np.unique(x))
            ]) < .33 * len(x):
            RP$x = ifelse(RP$x%%1 < .5, floor(RP$x), ceiling(RP$x))

        if Voronoi:
            abline(v = c(PART[ ,min(x), by=prior.quadrant]$V1,max(x)), lty = 3)
            if min_obs_stop:
                points(RP$x, RP$y, pch = 15, lwd = 2, col = "red")
            title(main = paste0("NNS Order = ", i), cex.main = 2)
        if min_obs_stop is False:
            RP = None
        return {
            "order": i,
            "dt": PART[],
            "regression.points": RP
        }
    return {}


__all__ = ["NNS_part"]



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd

# TODO TEST dydx
def dy_dx(x: pd.Series, y: pd.Series, eval_point: [float, int, str, None] = None, deriv_method: str = "FD") -> dict:
    r"""
    Partial Derivative dy/dx

    Returns the numerical partial derivate of \code{y} wrt \code{x} for a point of interest.

    @param x a numeric vector.
    @param y a numeric vector.
    @param eval_point numeric or ("overall"); \code{x} point to be evaluated.  Defaults to \code{(eval.point = median(x))}.  Set to \code{(eval.point = "overall")} to find an overall partial derivative estimate (1st derivative only).
    @param deriv.method method of derivative estimation, options: ("NNS", "FD"); Determines the partial derivative from the coefficient of the \link{NNS.reg} output when \code{(deriv.method = "NNS")} or generates a partial derivative using the finite difference method \code{(deriv.method = "FD")} (Defualt).
    @return Returns a list of both 1st and 2nd derivative:
    \itemize{
    \item{\code{dy.dx(...)$First}} the 1st derivative.
    \item{\code{dy.dx(...)$Second}} the 2nd derivative.
    }

    @note If a vector of derivatives is required, ensure \code{(deriv.method = "FD")}.
    @author Fred Viole, OVVO Financial Systems
    @references Viole, F. and Nawrocki, D. (2013) "Nonlinear Nonparametric Statistics: Using Partial Moments"
    \url{https://www.amazon.com/dp/1490523995}

    Vinod, H. and Viole, F. (2017) "Nonparametric Regression Using Clusters"
    \url{https://link.springer.com/article/10.1007/s10614-017-9713-5}

    @examples
    \dontrun{
    x <- seq(0, 2 * pi, pi / 100) ; y <-sin(x)
    dy.dx(x, y, eval.point = 1.75)

    # Vector of derivatives
    dy.dx(x, y, eval.point = c(1.75, 2.5), deriv.method = "FD")}
    @export
    """
    if eval_point is None:
        eval_point = x.median()
    order = None
    # TODO NNS_dep
    dep = NNS_dep(x, y)['Dependence']
    if(dep > 0.85):
        h = 0.01
    elif(dep > 0.5):
        h = 0.05
    else:
        h = 0.2

    if isinstance(x, pd.Series):
        x = pd.DataFrame(x, columns=[x.name])

    # TODO type of eval_point
    if (len(eval_point) > 1 and deriv_method == "NNS"):
        deriv_method = "FD"

    if isinstance(eval_point, str):
        # TODO NNS_reg
        ranges = NNS_reg(x, y, order = order, plot = False)['derivative']
        # TODO Convert
        ranges[ , interval := seq(1 : len(ranges['Coefficient']))]

        # TODO Convert
        range_weights = pd.DataFrame(x, 'interval' = findInterval(x, ranges[: , X_Lower_Range]))

        # TODO Convert
        ranges = ranges[interval in range_weights['interval'], ]

        # TODO Convert
        range_weights = range_weights[ , .N, by = 'interval']

        # TODO Convert
        range_weights = range_weights['N'] / range_weights['N'].sum()

        # TODO Convert
        return {
            "First": (ranges[:,Coefficient]*range_weights).sum()
        }

    original_eval_point_min = eval_point
    original_eval_point_max = eval_point

    # TODO Convert
    h_step = LPM_ratio(1, unlist(eval_point), x)
    # TODO Convert
    h_step = LPM_VaR(h_step + h, 1, x) - LPM_VaR(h_step - h, 1, x)

    # TODO Convert
    eval_point_min = original_eval_point_min - h_step

    # TODO Convert
    eval_point_max = h_step + original_eval_point_max

    # TODO Convert
    deriv_points = cbind(eval_point_min, eval_point, eval_point.max)

    # TODO Convert
    n = dim(deriv_points)[1]

    # TODO Convert
    run = eval_point_max - eval_point_min

    if (run==0).any():
        # TODO Convert
        z = which(run == 0)
        eval_point_max[z] = (abs((max(x) - min(x)) * h)) + eval_point[z]
        eval_point_max[z] = eval_point[z] - (abs((max(x) - min(x)) * h))
        run[z] = eval_point_max[z] - eval_point_min[z]

    # TODO Convert + NNS_reg
    reg_output = NNS_reg(x, y, plot = False, return_values = True, order = order, point_est = pd.Series(deriv_points))

    # TODO Convert
    estimates_min = reg_output['Point.est'][1:n]
    estimates_max = reg_output['Point.est'][(2*n+1):(3*n)]
    estimates = reg_output['Point.est'][(n+1):(2*n)]

    if deriv_method == "FD":
        # TODO Convert
        rise = estimates_max - estimates_min
        first_deriv =  rise / run
    else:
        # TODO Convert
        output = reg_output['derivative']
        if (len(output[: , Coefficient]) == 1):
            first_deriv = output[: , Coefficient]
        if((output[ , X_Upper_Range][which(eval.point < output[ , X_Upper_Range]) - 1][1]) < eval_point):
            first_deriv =  output[: , Coefficient][which(eval_point < output[: , X_Upper_Range])][1]
        else:
            first_deriv =  mean(c(output[: , Coefficient][which(eval_point < output[: , X_Upper_Range])][1], output[: , X_Lower_Range][which(eval_point < output[: , X_Upper_Range]) - 1][1]))

    # TODO Convert
    ## Second derivative form:
    # [f(x+h) - 2(f(x)) + f(x-h)] / h^2
    f_x__h = estimates_min
    two_f_x = 2 * estimates
    f_x_h = estimates_max
    second_deriv = (f_x_h - two_f_x + f_x__h) / (h_step ** 2)

    return {
        "First": first_deriv,
        "Second": second_deriv
    }


__all__ = [
    "dy_dx"
]



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from .LPM_UPM_VaR import LPM_VaR, UPM_VaR
from .Partial_Moments import LPM_ratio, UPM_ratio


def NNS_ANOVA_bin(
    control: [pd.Series, np.ndarray],
    treatment: [pd.Series, np.ndarray],
    mean_of_means: [float, None] = None,
    upper_25_target: [float, None] = None,
    lower_25_target: [float, None] = None,
    upper_125_target: [float, None] = None,
    lower_125_target: [float, None] = None,
    confidence_interval: [float, None] = None,
    tails: [str, None] = None,
    plot: bool = True,
    par=None,  # NO USE
) -> dict:
    if tails is None:
        tails = "both"
    if tails not in [None, "both", "left", "right"]:
        raise Exception("Tails must be None, both, left or right")

    if upper_25_target is None and lower_25_target is None:
        mean_of_means = np.mean([np.mean(control), np.mean(treatment)])
        upper_25_target = np.mean([UPM_VaR(0.25, 1, control), UPM_VaR(0.25, 1, treatment)])
        lower_25_target = np.mean([LPM_VaR(0.25, 1, control), LPM_VaR(0.25, 1, treatment)])
        upper_125_target = np.mean([UPM_VaR(0.125, 1, control), UPM_VaR(0.125, 1, treatment)])
        lower_125_target = np.mean([LPM_VaR(0.125, 1, control), LPM_VaR(0.125, 1, treatment)])

    # Continuous CDF for each variable from Mean of Means
    LPM_ratio_1 = LPM_ratio(1, mean_of_means, control)
    LPM_ratio_2 = LPM_ratio(1, mean_of_means, treatment)

    Upper_25_ratio_1 = UPM_ratio(1, upper_25_target, control)
    Upper_25_ratio_2 = UPM_ratio(1, upper_25_target, treatment)
    # Upper_25_ratio = np.mean([Upper_25_ratio_1, Upper_25_ratio_2])

    Lower_25_ratio_1 = LPM_ratio(1, lower_25_target, control)
    Lower_25_ratio_2 = LPM_ratio(1, lower_25_target, treatment)
    # Lower_25_ratio = np.mean([Lower_25_ratio_1, Lower_25_ratio_2])

    Upper_125_ratio_1 = UPM_ratio(1, upper_125_target, control)
    Upper_125_ratio_2 = UPM_ratio(1, upper_125_target, treatment)
    # Upper_125_ratio = np.mean([Upper_125_ratio_1, Upper_125_ratio_2])

    Lower_125_ratio_1 = LPM_ratio(1, lower_125_target, control)
    Lower_125_ratio_2 = LPM_ratio(1, lower_125_target, treatment)
    # Lower_125_ratio = np.mean([Lower_125_ratio_1, Lower_125_ratio_2])

    # Continuous CDF Deviation from 0.5
    MAD_CDF = min(0.5, max(abs(LPM_ratio_1 - 0.5), abs(LPM_ratio_2 - 0.5)))
    upper_25_CDF = min(0.25, max(abs(Upper_25_ratio_1 - 0.25), abs(Upper_25_ratio_2 - 0.25)))
    lower_25_CDF = min(0.25, max(abs(Lower_25_ratio_1 - 0.25), abs(Lower_25_ratio_2 - 0.25)))
    upper_125_CDF = min(0.125, max(abs(Upper_125_ratio_1 - 0.125), abs(Upper_125_ratio_2 - 0.125)))
    lower_125_CDF = min(0.125, max(abs(Lower_125_ratio_1 - 0.125), abs(Lower_125_ratio_2 - 0.125)))

    # Certainty associated with samples
    NNS_ANOVA_rho = (
        np.sum(
            [
                ((0.5 - MAD_CDF) ** 2) / 0.25,
                0.5 * (((0.25 - upper_25_CDF) ** 2) / 0.25 ** 2),
                0.5 * (((0.25 - lower_25_CDF) ** 2) / 0.25 ** 2),
                0.25 * (((0.125 - upper_125_CDF) ** 2) / 0.125 ** 2),
                0.25 * (((0.125 - lower_125_CDF) ** 2) / 0.125 ** 2),
            ]
        )
        / 2.5
    )

    pop_adjustment = ((len(control) + len(treatment) - 2) / (len(control) + len(treatment))) ** 2

    # Graphs
    if plot:
        plt.title("NNS ANOVA and Effect Side")
        plt.boxplot(
            [control, treatment],
            labels=["Control", "Treatment"],
            vert=False,
        )
        plt.xlabel("Means")
        plt.axvline(mean_of_means, color="red", linewidth=4, label="Grand Mean")
        # if par is None:
        #    original_par = par(no.readonly = TRUE)
        # else:
        #    original_par = par
        # boxplot(list(control, treatment), las = 2, names = c("Control", "Treatment"), xlab = "Means", horizontal = TRUE, main = "NNS ANOVA and Effect Size", col = c("grey", "white"), cex.axis = 0.75)
        # For ANOVA Visualization
        # abline(v = mean_of_means, col = "red", lwd = 4)
        # mtext("Grand Mean", side = 3, col = "red", at = mean_of_means)

    if confidence_interval is None:
        return {
            "Control Mean": np.mean(control),
            "Treatment Mean": np.mean(treatment),
            "Grand Mean": mean_of_means,
            "Control CDF": LPM_ratio_1,
            "Treatment CDF": LPM_ratio_2,
            "Certainty": min(1, NNS_ANOVA_rho * pop_adjustment),
        }
    else:
        # Upper end of CDF confidence interval for control mean
        Upper_Bound_Effect, Lower_Bound_Effect, CI = None, None, 0
        if tails == "both":
            CI = confidence_interval + (1 - confidence_interval) / 2
        elif tails in ["left", "right"]:
            CI = confidence_interval
        a = UPM_VaR(1 - CI, 1, control)
        b = np.mean(control)
        if plot:
            if tails in ["both", "right"]:
                plt.axvline(max(a, b), color="green", linewidth=4, label="mu+", linestyle=":")
                # abline(v = max(a, b), col = "green", lwd = 4, lty = 3)
                # text(max(a, b), pos = 2, 0.75, "mu+", col = "green")
                # text(max(a, b), pos = 4, 0.75, paste0((1 - CI) * 100, "% --->"), col = "green")}

        # Lower end of CDF confidence interval for control mean
        c = LPM_VaR(1 - CI, 1, control)
        d = np.mean(control)

        if plot:
            if tails in ["both", "left"]:
                plt.axvline(min(c, d), color="blue", linewidth=4, label="mu-", linestyle=":")
                # abline(v = min(c, d), col = "blue", lwd = 4, lty = 3)
                # text(min(c, d), pos = 4, 0.75, "mu-", col = "blue")
                # text(min(c, d), pos=2, 0.75, paste0( "<--- ", (1 - CI) * 100, "%"), col = 'blue')}
            # par(original.par)

        # Effect Size Lower Bound
        if tails in ["both", "right"]:
            Lower_Bound_Effect = np.mean(treatment) - max(a, b)
        elif tails == "left":
            Lower_Bound_Effect = np.mean(treatment) - max(c, d)

        # Effect Size Upper Bound
        if tails in ["both", "left"]:
            Upper_Bound_Effect = np.mean(treatment) - min(c, d)
        elif tails == "right":
            Upper_Bound_Effect = np.mean(treatment) - min(a, b)

        # Certainty Statistic and Effect Size Given Confidence Interval
        return {
            "Control Mean": np.mean(control),
            "Treatment Mean": np.mean(treatment),
            "Grand Mean": mean_of_means,
            "Control CDF": LPM_ratio_1,
            "Treatment CDF": LPM_ratio_2,
            "Certainty": min(1, NNS_ANOVA_rho * pop_adjustment),
            "Lower Bound Effect": Lower_Bound_Effect,
            "Upper Bound Effect": Upper_Bound_Effect,
        }


__all__ = ["NNS_ANOVA_bin"]



# -*- coding: utf-8 -*-
import pandas as pd


def NNS_term_matrix(x: pd.DataFrame, oos: [None, pd.Series] = None, names: bool = False) -> dict:
    r"""NNS Term Matrix

    Generates a term matrix for text classification use in \link{NNS.reg}.

    @param x mixed data.frame; character/numeric; A two column dataset should be used.  Concatenate text from original sources to comply with format.  Also note the possibility of factors in \code{"DV"}, so \code{"as.numeric(as.character(...))"} is used to avoid issues.
    @param oos mixed data.frame; character/numeric; Out-of-sample text dataset to be classified.
    @param names logical; Column names for \code{"IV"} and \code{"oos"}.  Defaults to FALSE.
    @return Returns the text as independent variables \code{"IV"} and the classification as the dependent variable \code{"DV"}.  Out-of-sample independent variables are returned with \code{"OOS"}.
    @references Viole, F. and Nawrocki, D. (2013) "Nonlinear Nonparametric Statistics: Using Partial Moments"
    \url{https://www.amazon.com/dp/1490523995/ref=cm_sw_su_dp}
    @examples
    x <- data.frame(cbind(c("sunny", "rainy"), c(1, -1)))
    NNS.term.matrix(x)

    ### Concatenate Text with space separator, cbind with "DV"
    x <- data.frame(cbind(c("sunny", "rainy"), c("windy", "cloudy"), c(1, -1)))
    x <- data.frame(cbind(paste(x[ , 1], x[ , 2], sep = " "), as.numeric(as.character(x[ , 3]))))
    NNS.term.matrix(x)


    ### NYT Example
    \dontrun{
    require(RTextTools)
    data(NYTimes)

    ### Concatenate Columns 3 and 4 containing text, with column 5 as DV
    NYT <- data.frame(cbind(paste(NYTimes[ , 3], NYTimes[ , 4], sep = " "),
                          as.numeric(as.character(NYTimes[ , 5]))))
    NNS.term.matrix(NYT)}
    @export
    """

    def count_unique(x1) -> list:
        return [x1.count(ii) for ii in unique_vocab]

    p = 0 if oos is None else len(oos)
    n = len(x)
    # Remove commas, etc.
    x.iloc[:, 0] = x.iloc[:, 0].astype(str)
    for i in [",", ";", ":", "'s", " . "]:
        x.iloc[:, 0] = x.iloc[:, 0].str.replace(i, "")

    unique_vocab = list(x.iloc[:, 0].str.split(" ", expand=True).stack().unique())

    # Sub with a longer .csv to be called to reduce IVs
    prepositions = [
        "a",
        "in",
        "of",
        "our",
        "the",
        "is",
        "for",
        "with",
        "we",
        "this",
        "it",
        "but",
        "was",
        "at",
        "to",
        "on",
        "aboard",
        "aside",
        "by",
        "means",
        "spite",
        "about",
        "as",
        "concerning",
        "instead",
        "above",
        "at",
        "considering",
        "into",
        "according",
        "atop",
        "despite",
        "view",
        "across",
        "because",
        "during",
        "near",
        "like",
        "across",
        "after",
        "against",
        "ahead",
        "along",
        "alongside",
        "amid",
        "among",
        "apart",
        "around",
        "out",
        "outside",
        "over",
        "owing",
        "past",
        "prior",
        "before",
        "behind",
        "below",
        "beneath",
        "beside",
        "besides",
        "between",
        "beyond",
        "regarding",
        "round",
        "since",
        "through",
        "throughout",
        "till",
        "down",
        "except",
        "from",
        "addition",
        "back",
        "front",
        "place",
        "regard",
        "inside",
        "together",
        "toward",
        "under",
        "underneath",
        "until",
        "nearby",
        "next",
        "off",
        "account",
        "onto",
        "top",
        "opposite",
        "out",
        "unto",
        "up",
        "within",
        "without",
        "what",
    ]

    # Remove prepositions
    preps = [i for i in unique_vocab if i in prepositions]

    if len(preps) > 0:
        unique_vocab = [i for i in unique_vocab if i not in prepositions]

    if oos is not None:
        oos_preps = [i for i in oos if i in prepositions]
        if len(oos_preps) > 0:
            oos = [i for i in oos if i not in prepositions]
        oos = pd.Series(oos)

    NNS_TM = x.iloc[:, 0].apply(count_unique).apply(pd.Series)
    # NNS_TM = t(
    #    sapply(
    #        1 : length(x[ , 1]),
    #        function(i) as.integer(
    #            tryCatch(
    #                stringr::str_count(x[i, 1], unique.vocab),
    #                error = function (e) 0
    #            )
    #        )
    #    )
    # )
    if names:
        NNS_TM.columns = unique_vocab

    if oos is not None:
        OOS_TM = oos.apply(count_unique).apply(pd.Series)
        # OOS_TM = t(
        #    sapply(
        #        1 : length(oos),
        #        function(i) stringr::str_count(oos[i], unique.vocab)
        #    )
        # )
        if names:
            OOS_TM.columns = unique_vocab

        return {"IV": NNS_TM, "DV": x.iloc[:, 1].values, "OOS": OOS_TM}

    return {"IV": NNS_TM, "DV": x.iloc[:, 1].values}


__all__ = ["NNS_term_matrix"]



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from .Partial_Moments import LPM

# TODO: test / matplotlib graph
def NNS_SSD(x: pd.Series, y: pd.Series, use_plot: bool = True) -> str:
    r"""
    NNS SSD Test

    Bi-directional test of second degree stochastic dominance using lower partial moments.

    @param x a numeric vector.
    @param y a numeric vector.
    @return Returns one of the following SSD results: \code{"X SSD Y"}, \code{"Y SSD X"}, or \code{"NO SSD EXISTS"}.
    @author Fred Viole, OVVO Financial Systems
    @references Viole, F. and Nawrocki, D. (2016) "LPM Density Functions for the Computation of the SD Efficient Set." Journal of Mathematical Finance, 6, 105-126. \url{http://www.scirp.org/Journal/PaperInformation.aspx?PaperID=63817}.
    @examples
    set.seed(123)
    x <- rnorm(100) ; y <- rnorm(100)
    NNS.SSD(x, y)
    @export
    """
    x_sort = np.sort(x)
    y_sort = np.sort(y)

    Combined_sort = np.unique(np.append(x_sort, y_sort))
    LPM_x_sort = LPM(1, Combined_sort, x)
    LPM_y_sort = LPM(1, Combined_sort, y)
    x_ssd_y = np.any(LPM_x_sort > LPM_y_sort)
    y_ssd_x = np.any(LPM_y_sort > LPM_x_sort)
    if use_plot:
        plt.title("SSD")
        plt.ylabel("Area of Cumulative Distribution")
        plt.plot(Combined_sort, LPM_x_sort, label="<Combined Sort> vs <LPM X Sort>")
        plt.plot(Combined_sort, LPM_y_sort, label="<Combined Sort> vs <LPM Y Sort>")
        plt.legend()
        # plot(
        #    Combined_sort,
        #    LPM_x_sort,
        #    type = "l",
        #    lwd = 3,
        #    col = "red",
        #    main = "SSD",
        #    ylab = "Area of Cumulative Distribution",
        #    ylim = c(min(c(LPM_y_sort, LPM_x_sort)), max(c(LPM_y_sort, LPM_x_sort)))
        # )
        # lines(Combined_sort, LPM_y_sort, type = "l", lwd = 3,col = "blue")
        # legend("topleft", c("X", "Y"), lwd = 10, col = c("red", "blue"))

    if (
        (not x_ssd_y)
        and (x_sort[0] >= y_sort[0])
        and (x.mean() >= y.mean())
        and (not np.equal(LPM_x_sort, LPM_y_sort).all())
    ):
        return "X SSD Y"
    if (
        (not y_ssd_x)
        and (y_sort[0] >= x_sort[0])
        and (y.mean() >= x.mean())
        and (not np.equal(LPM_x_sort, LPM_y_sort).all())
    ):
        return "Y SSD X"
    return "NO SSD EXISTS"


__all__ = [
    "NNS_SSD",
]



# -*- coding: utf-8 -*-
import numpy as np
from typing import Callable
import scipy.optimize
import scipy.linalg


def NNS_diff(
    f: Callable,
    point: float,
    h: float = 0.1,
    tol: float = 1e-10,
    digits: int = 12,  # not in use
    print_trace: bool = False,
) -> dict:
    r"""NNS Numerical Differentiation

    Determines numerical derivative of a given univariate function using projected secant lines on the y-axis.  These projected points infer finite steps \code{h}, in the finite step method.

    @param f an expression or call or a formula with no lhs.
    @param point numeric; Point to be evaluated for derivative of a given function \code{f}.
    @param h numeric [0, ...]; Initial step for secant projection.  Defaults to \code{(h = 0.1)}.
    @param tol numeric; Sets the tolerance for the stopping condition of the inferred \code{h}.  Defaults to \code{(tol = 1e-10)}.
    @param digits numeric; Sets the number of digits specification of the output.  Defaults to \code{(digits = 12)}.
    @param print.trace logical; \code{FALSE} (default) Displays each iteration, lower y-intercept, upper y-intercept and inferred \code{h}.
    @return Returns a matrix of values, intercepts, derivatives, inferred step sizes for multiple methods of estimation.
    @author Fred Viole, OVVO Financial Systems
    @references Viole, F. and Nawrocki, D. (2013) "Nonlinear Nonparametric Statistics: Using Partial Moments"
    \url{https://www.amazon.com/dp/1490523995/ref=cm_sw_su_dp}
    @examples
    f <- function(x) sin(x) / x
    NNS.diff(f, 4.1)
    @export

    """

    def Finite_step(p: float, h: float) -> dict:
        f_x = f(p)
        f_x_h_min = f(p - h)
        f_x_h_pos = f(p + h)

        neg_step = (f_x - f_x_h_min) / h
        pos_step = (f_x_h_pos - f_x) / h

        return {
            "f(x-h)": neg_step,
            "f(x+h)": pos_step,
            "next": np.mean([neg_step, pos_step]),
        }

    # Bs <- numeric()
    # Bl <- numeric()
    # Bu <- numeric()
    Bs, Bl, Bu = {}, {}, {}
    ### Step 1 initalize the boundaries for B

    ### Initial step size h
    f_x = f(point)
    f_x_h = f(point - h)

    ### Y = mX + B
    Y = f_x
    m = (f_x - f_x_h) / h
    mX = ((f_x - f_x_h) / h) * point
    B = f_x - ((f_x - f_x_h) / h) * point
    ### Initial interval for B given inputted h-step-value

    f_x_h_lower = f(point - h)
    f_x_h_upper = f(point + h)

    B1 = f_x - ((f_x - f_x_h_lower) / h) * point
    B2 = f_x - ((f_x_h_upper - f_x) / h) * point

    low_B = min(B1, B2)
    high_B = max(B1, B2)

    lower_B = low_B
    upper_B = high_B

    ## Return "Derivative Does Not Exist" if lower.B and upper.B are identical to 20 digits
    if lower_B == upper_B:
        # TODO graph
        # original_par = par(no.readonly = TRUE)
        # par(mfrow = c(1, 2))
        # plot(f, xlim = c(point - (100 * h), point + (100 * h)), col = 'blue', ylab = 'f(x)')
        # points(point, f.x, pch = 19, col = 'red')
        # plot(f, xlim = c(point - 1, point + 1), col = 'blue', ylab = 'f(x)')
        # points(point, f.x, pch = 19, col = 'red')
        # par(original.par)
        raise Exception("Derivative Does Not Exist")

    new_B = np.mean([lower_B, upper_B])
    i = 1
    while i >= 1:
        Bl[i] = lower_B
        Bu[i] = upper_B
        # TODO: understand function(x)
        # new.f <- function(x) - f.x + ((f.x - f(point - x)) / x) * point + new.B
        ###  SOLVE FOR h, we just need the negative or positive sign from the tested B
        # inferred_h = uniroot(new_f, c(-2 * h, 2 * h), extendInt="yes")["root"]

        inferred_h = scipy.optimize.root_scalar(
            f=lambda x: -f_x + ((f_x - f(point - x)) / x) * point + new_B,
            # method='bisect',
            bracket=[-2 * h, 2 * h],
        ).root

        if print_trace:
            print("Iteration", i, "h", inferred_h, "Lower B", lower_B, "Upper B", upper_B)

        Bs[i] = new_B

        ## Stop when the inferred h is within the tolerance level
        if abs(inferred_h) < tol:
            final_B = np.mean([upper_B, lower_B])
            # TODO: scipy solver
            slope = scipy.linalg.solve(point, f_x - final_B)[0]
            z = complex(real=point, imag=inferred_h)

            # TODO: graph
            # original_par <- par(no.readonly = TRUE)
            # par(mfrow=c(1, 3))
            ## Plot #1
            # plot(f, xlim = c(min(c(point - (100 * h), point + (100 * h)), 0), max(c(point - (100 * h), point + (100 * h)), 0)), col = 'azure4', ylab = 'f(x)', lwd = 2, ylim = c(min(c(min(c(B1, B2)), min(na.omit(f((point - (100 * h)) : (point + (100 * h))))))), max(c(max(na.omit(f((point - (100 * h)) : (point + (100 * h))))), max(c(B1, B2))))), main = 'f(x) and initial y-intercept range')
            # abline(h = 0, v = 0, col = 'grey')
            # points(point, f.x, pch = 19, col = 'green')
            # points(point - h, f.x.h.lower, col = ifelse(B1 == high.B, 'blue', 'red'), pch = 19)
            # points(point + h, f.x.h.upper, col = ifelse(B1 == high.B, 'red', 'blue'), pch = 19)
            # points(x = rep(0, 2), y = c(B1, B2), col = c(ifelse(B1 == high.B, 'blue', 'red'), ifelse(B1 == high.B, 'red', 'blue')), pch = 1)
            # segments(0, B1, point - h, f.x.h.lower, col = ifelse(B1 == high.B, 'blue','red'), lty = 2)
            # segments(0, B2, point + h, f.x.h.upper, col = ifelse(B1 == high.B, 'red','blue'), lty = 2)

            ### Plot #2
            # plot(f, col = 'azure4', ylab = 'f(x)', lwd = 3, main = 'f(x) narrowed range and secant lines', xlim = c(min(c(point - h, point + h,  0)), max(c(point + h,point - h, 0))), ylim= c(min(c(B1, B2, f.x.h.lower, f.x.h.upper)), max(c(B1, B2, f.x.h.lower, f.x.h.upper))))
            # abline(h = 0, v = 0, col = 'grey')
            # points(point,f.x, pch = 19, col = 'red')
            # points(point - h, f.x.h.lower, col = ifelse(B1 == high.B, 'blue', 'red'), pch = 19)
            # points(point + h, f.x.h.upper, col = ifelse(B1 == high.B, 'red', 'blue'), pch = 19)
            # points(point, f.x, pch = 19, col = 'green')
            # segments(0, B1, point - h, f.x.h.lower, col = ifelse(B1 == high.B, 'blue', 'red'), lty = 2)
            # segments(0, B2, point + h, f.x.h.upper, col = ifelse(B1 == high.B, 'red', 'blue'), lty = 2)
            # points(x = rep(0, 2), y = c(B1, B2), col = c(ifelse(B1 == high.B, 'blue', 'red'), ifelse(B1 == high.B, 'red', 'blue')), pch = 1)

            ## Plot #3
            # plot(Bs, ylim = c(min(c(Bl, Bu)), max(c(Bl, Bu))), xlab = "Iterations", ylab = "y-inetercept", col = 'green', pch = 19, main = 'Iterated range of y-intercept')
            # points(Bl, col = 'red', ylab = '')
            # points(Bu, col = 'blue', ylab = '')
            # legend('topright', c("Upper y-intercept", "Lower y-intercept", "Mean y-intercept"), col = c('blue', 'red', 'green'), pch = c(1, 1, 19), bty = 'n')
            # par(original.par)
            ret = Finite_step(point, h)
            ret.update(
                {
                    "Value of f(x) at point": f(point),
                    "Final y-intercept (B)": final_B,
                    "DERIVATIVE": slope,
                    "Inferred.h": inferred_h,
                    "iterations": i,
                    "Averaged Finite Step Initial h": Finite_step(point, h)["next"],
                    "Inferred h": Finite_step(point, inferred_h),
                    "Inferred h Averaged Finite Step": Finite_step(point, inferred_h)["next"],
                    "Complex Step Derivative (Initial h)": complex(f(z)).imag / complex(z).imag,
                }
            )
            return ret

        ## NARROW THE RANGE OF B BASED ON SIGN OF INFERRED.H
        if B1 == high_B:
            if np.sign(inferred_h) < 0:
                lower_B = new_B
                upper_B = upper_B
            else:
                upper_B = new_B
                lower_B = lower_B
        else:
            if np.sign(inferred_h) < 0:
                lower_B = lower_B
                upper_B = new_B
            else:
                upper_B = upper_B
                lower_B = new_B

        new_B = np.mean([lower_B, upper_B])
        i += 1


__all__ = ["NNS_diff"]



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
from .Partial_Moments import LPM_ratio
import matplotlib.pyplot as plt

# TODO: test / matplotlib
def NNS_FSD(x: pd.Series, y: pd.Series, type_cdf: str = "discrete", use_plot: bool = True) -> str:
    r"""
    NNS FSD Test

    Bi-directional test of first degree stochastic dominance using lower partial moments.

    @param x a numeric vector.
    @param y a numeric vector.
    @param type options: ("discrete", "continuous"); \code{"discrete"} (default) selects the type of CDF.
    @return Returns one of the following FSD results: \code{"X FSD Y"}, \code{"Y FSD X"}, or \code{"NO FSD EXISTS"}.
    @author Fred Viole, OVVO Financial Systems
    @references Viole, F. and Nawrocki, D. (2016) "LPM Density Functions for the Computation of the SD Efficient Set." Journal of Mathematical Finance, 6, 105-126. \url{http://www.scirp.org/Journal/PaperInformation.aspx?PaperID=63817}.

    Viole, F. (2017) "A Note on Stochastic Dominance." \url{https://ssrn.com/abstract=3002675}.
    @examples
    set.seed(123)
    x <- rnorm(100) ; y <- rnorm(100)
    NNS.FSD(x, y)
    @export
    """
    type_cdf = type_cdf.lower()
    if type_cdf not in ["discrete", "continuous"]:
        raise Exception("type needs to be either 'discrete' or 'continuous'")

    x_sort = np.sort(x)
    y_sort = np.sort(y)
    Combined_sort = np.unique(np.append(x_sort, y_sort))
    ## Indicator function ***for all values of x and y*** as the continuous CDF target
    degree = 0 if type_cdf == "discrete" else 1
    LPM_x_sort = LPM_ratio(degree, Combined_sort, x)
    LPM_y_sort = LPM_ratio(degree, Combined_sort, y)

    x_fsd_y = np.any(LPM_x_sort > LPM_y_sort)
    y_fsd_x = np.any(LPM_y_sort > LPM_x_sort)

    if use_plot:
        plt.title("FSD")
        plt.ylabel("Area of Cumulative Distribution")
        plt.plot(Combined_sort, LPM_x_sort, label="<Combined Sort> vs <LPM X Sort>")
        plt.plot(Combined_sort, LPM_y_sort, label="<Combined Sort> vs <LPM Y Sort>")
        plt.legend()
        # plot(
        #    Combined_sort,
        #    LPM_x_sort,
        #    type="l",
        #    lwd=3,
        #    col="red",
        #    main="FSD",
        #    ylab="Probability of Cumulative Distribution",
        #    ylim=c(0, 1),
        # )
        # lines(Combined_sort, LPM_y_sort, type="l", lwd=3, col="blue")
        # legend("topleft", c("X", "Y"), lwd=10, col=c("red", "blue"))

    ## Verification of ***0 instances*** of CDFx > CDFy, and conversely of CDFy > CDFx
    if (not x_fsd_y) and (x_sort[0] >= y_sort[0]) and (not np.equal(LPM_x_sort, LPM_y_sort).all()):
        return "X FSD Y"
    if (not y_fsd_x) and (y_sort[0] >= x_sort[0]) and (not np.equal(LPM_x_sort, LPM_y_sort).all()):
        return "Y FSD X"
    return "NO FSD EXISTS"


__all__ = ["NNS_FSD"]



# -*- coding: utf-8 -*-
# https://raw.githubusercontent.com/pypa/sampleproject/master/setup.py
# Always prefer setuptools over distutils
import pathlib

from setuptools import find_packages, setup

here = pathlib.Path(__file__).parent.resolve()
# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.
setup(
    name="NNS",
    version="0.1.10",
    description="Nonlinear nonparametric statistics using partial moments",
    long_description=long_description,
    long_description_content_type="text/markdown",  # Optional (see note above)
    author="Fred Viole, Roberto Spadim",
    author_email="ovvo.financial.systems@gmail.com",
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        # Indicate who your project is intended for
        # TODO: 'Intended Audience :: Developers',
        # Pick your license as you wish
        # TODO: 'License :: OSI Approved :: MIT License',
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate you support Python 3. These classifiers are *not*
        # checked by 'pip install'. See instead 'python_requires' below.
        "Programming Language :: Python :: 3.7",
    ],
    keywords="Nonlinear nonparametric regression classification clustering",
    # When your source code is in a subdirectory under the project root, e.g.
    # `src/`, it is necessary to specify the `package_dir` argument.
    # package_dir={'': 'src'},  # Optional
    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    # packages=find_packages(where='src'),
    packages=find_packages(),
    # Specify which Python versions you support. In contrast to the
    # 'Programming Language' classifiers above, 'pip install' will check this
    # and refuse to install the project if the version does not match. See
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    python_requires=">=3.7",
    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        "numpy>=1.19.0",
        "pandas>=1.0.5",
        "scikit_learn>=0.23.1",
        "KDEpy>=1.1.0",
    ],
    extras_require={
        "test": ["pytest-cov==2.10.0", "pytest==5.4.3"],
    },
    url="https://github.com/OVVO-Financial/NNS-Python/",
    # https://medium.com/@joel.barmettler/how-to-upload-your-python-package-to-pypi-65edc5fe9c56
    # build:  python setup.py sdist
    # upload: twine upload dist/*
    download_url="https://github.com/OVVO-Financial/NNS-Python/archive/refs/tags/v_0110.tar.gz",
)



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from .Partial_Moments import PM_matrix

_tmp = Axes3D, Poly3DCollection  # just o avoid import lint errors
del _tmp

# https://stackoverflow.com/questions/49277753/python-matplotlib-plotting-cuboids
def cuboid_data(o, size=(1, 1, 1)):
    # code taken from
    # https://stackoverflow.com/a/35978146/4124317
    # suppose axis direction: x: to left; y: to inside; z: to upper
    # get the length, width, and height
    l, w, h = size
    x = [
        [o[0], o[0] + l, o[0] + l, o[0], o[0]],
        [o[0], o[0] + l, o[0] + l, o[0], o[0]],
        [o[0], o[0] + l, o[0] + l, o[0], o[0]],
        [o[0], o[0] + l, o[0] + l, o[0], o[0]],
    ]
    y = [
        [o[1], o[1], o[1] + w, o[1] + w, o[1]],
        [o[1], o[1], o[1] + w, o[1] + w, o[1]],
        [o[1], o[1], o[1], o[1], o[1]],
        [o[1] + w, o[1] + w, o[1] + w, o[1] + w, o[1] + w],
    ]
    z = [
        [o[2], o[2], o[2], o[2], o[2]],
        [o[2] + h, o[2] + h, o[2] + h, o[2] + h, o[2] + h],
        [o[2], o[2], o[2] + h, o[2] + h, o[2]],
        [o[2], o[2], o[2] + h, o[2] + h, o[2]],
    ]
    return np.array(x), np.array(y), np.array(z)


def plotCubeAt(pos=(0, 0, 0), size=(1, 1, 1), ax=None, **kwargs):
    # Plotting a cube element at position pos
    if ax is not None:
        X, Y, Z = cuboid_data(pos, size)
        ax.plot_surface(X, Y, Z, rstride=1, cstride=1, **kwargs)


def NNS_copula(
    x: [pd.DataFrame, pd.Series, np.ndarray],
    continuous: bool = True,
    plot: bool = True,
    independence_overlay: bool = False,
) -> [float, None]:
    r"""NNS Co-Partial Moments Higher Dimension Dependence

    Determines higher dimension dependence coefficients based on co-partial moment matrices ratios.

    @param x a numeric matrix or data frame.
    @param continuous logical; \code{TRUE} (default) Generates a continuous measure using degree 1 \link{PM.matrix}, while discrete \code{FALSE} uses degree 0 \link{PM.matrix}.
    @param plot logical; \code{FALSE} (default) Generates a 3d scatter plot with regression points using \link{plot3d}.
    @param independence_overlay logical; \code{FALSE} (default) Creates and overlays independent \link{Co.LPM} and \link{Co.UPM} regions to visually reference the difference in dependence from the data.frame of variables being analyzed.  Under independence, the light green and red shaded areas would be occupied by green and red data points respectively.

    @return Returns a multivariate dependence value [0,1].

    @author Fred Viole, OVVO Financial Systems
    @references Viole, F. (2016) "Beyond Correlation: Using the Elements of Variance for Conditional Means and Probabilities"  \url{https://www.ssrn.com/abstract=2745308}.
    @examples
    set.seed(123)
    x <- rnorm(1000) ; y <- rnorm(1000) ; z <- rnorm(1000)
    A <- data.frame(x, y, z)
    NNS.copula(A, plot = TRUE, independence.overlay = TRUE)
    @export
    """

    if np.any(np.isnan(x)):
        raise Exception("You have some missing values, please address.")
    if isinstance(x, pd.Series):
        x = x.to_frame()
    if isinstance(x, np.ndarray) and len(x.shape) == 1:
        x = x.reshape(-1, 1)  # series to matrix like

    A = x
    n = A.shape[1]
    colnames = x.columns if isinstance(x, pd.DataFrame) else list(range(x.shape[1]))
    # l = A.shape[0]

    # if(is.null(colnames(A))){
    #    colnames.list <- list()
    #    for(i in 1 : n){
    #        colnames.list[i] <- paste0("Var ", i)
    #    }
    #    colnames(A) <- c(colnames.list)
    # }

    # if(continuous) degree <- 1 else degree <- 0
    degree = 1 if continuous else 0

    # Generate partial moment matrices
    # pm_cov <- PM.matrix(degree, degree, variable = x, pop.adj = TRUE)
    pm_cov = PM_matrix(degree, degree, variable=x, pop_adj=True)

    # Isolate the upper triangles from each of the partial moment matrices
    Co_pm = np.sum(np.triu(pm_cov["cupm"], 1)) + np.sum(np.triu(pm_cov["clpm"], 1))
    D_pm = np.sum(np.triu(pm_cov["dupm"], 1)) + np.sum(np.triu(pm_cov["dlpm"], 1))
    if plot and n == 3:
        fig = plt.figure(figsize=(4, 4))
        ax = fig.add_subplot(111, projection="3d")
        B = A.values if isinstance(A, pd.DataFrame) else A
        tmp_color_1l = B[:, 0] <= np.mean(B[:, 0])
        tmp_color_2l = B[:, 1] <= np.mean(B[:, 1])
        tmp_color_3l = B[:, 2] <= np.mean(B[:, 2])
        tmp_color_1g = B[:, 0] > np.mean(B[:, 0])
        tmp_color_2g = B[:, 1] > np.mean(B[:, 1])
        tmp_color_3g = B[:, 2] > np.mean(B[:, 2])
        tmp_color = [
            "red"
            if (tmp_color_1l[i] and tmp_color_2l[i] and tmp_color_3l[i])
            else (
                "green"
                if (tmp_color_1g[i] and tmp_color_2g[i] and tmp_color_3g[i])
                else "steelblue"
            )
            for i in range(len(tmp_color_1l))
        ]
        del tmp_color_1g, tmp_color_1l, tmp_color_2g, tmp_color_2l, tmp_color_3g, tmp_color_3l
        ax.scatter(
            B[:, 0],
            B[:, 1],
            B[:, 2],
            c=tmp_color,
        )
        ax.set_xlabel(colnames[0])
        ax.set_ylabel(colnames[1])
        ax.set_zlabel(colnames[2])
        if independence_overlay:
            positions = [
                (np.min(B[:, 0]), np.min(B[:, 1]), np.min(B[:, 2])),
                (np.mean(B[:, 0]), np.mean(B[:, 1]), np.mean(B[:, 2])),
            ]
            sizes = [
                (
                    np.mean(B[:, 0]) - np.min(B[:, 0]),
                    np.mean(B[:, 1]) - np.min(B[:, 1]),
                    np.mean(B[:, 2]) - np.min(B[:, 2]),
                ),
                (
                    np.max(B[:, 0]) - np.mean(B[:, 0]),
                    np.max(B[:, 1]) - np.mean(B[:, 1]),
                    np.max(B[:, 2]) - np.mean(B[:, 2]),
                ),
            ]
            colors = ["red", "green"]
            for p, s, c in zip(positions, sizes, colors):
                plotCubeAt(pos=p, size=s, ax=ax, color=c, alpha=0.25)

    if np.any(np.isnan(Co_pm)) or Co_pm is None:
        Co_pm = 0
    if np.any(np.isnan(D_pm)) or D_pm is None:
        D_pm = 0
    if Co_pm == D_pm:
        return 0
    elif Co_pm == 0 or D_pm == 0:
        return 1
    elif Co_pm < D_pm:
        return 1 - Co_pm / D_pm
    elif Co_pm > D_pm:
        return 1 - D_pm / Co_pm
    return None


__all__ = ["cuboid_data", "plotCubeAt", "NNS_copula"]



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd

from .Copula import NNS_copula
from .Partial_Moments import LPM, UPM
from .Partition_Map import NNS_part


def NNS_dep(x: [np.ndarray, pd.Series], y: [np.ndarray, pd.Series, None]=None, asym: bool=False,p_value: bool=False,print_map: bool=False) -> dict:
    """
    NNS Dependence
    
    Returns the dependence and nonlinear correlation between two variables based on higher order partial moment matrices measured by frequency or area.
    
    @param x a numeric vector, matrix or data frame.
    @param y \code{NULL} (default) or a numeric vector with compatible dimensions to \code{x}.
    @param asym logical; \code{FALSE} (default) Allows for asymmetrical dependencies.
    @param p_value logical; \code{FALSE} (default) Generates 100 independent random permutations to test results against and plots 95 percent confidence intervals along with all results.
    @param print_map logical; \code{FALSE} (default) Plots quadrant means, or p-value replicates.
    @return Returns the bi-variate \code{"Correlation"} and \code{"Dependence"} or correlation / dependence matrix for matrix input.
    
    @note
    \code{NNS.cor} has been deprecated \code{(NNS >= 0.5.4)} and can be called via \code{NNS.dep}.
    
    @author Fred Viole, OVVO Financial Systems
    @references Viole, F. and Nawrocki, D. (2013) "Nonlinear Nonparametric Statistics: Using Partial Moments"
    \url{https://www.amazon.com/dp/1490523995/ref=cm_sw_su_dp}
    @examples
    \dontrun{
    set.seed(123)
    x <- rnorm(100) ; y <- rnorm(100)
    NNS.dep(x, y)
    
    ## Correlation / Dependence Matrix
    x <- rnorm(100) ; y <- rnorm(100) ; z <- rnorm(100)
    B <- cbind(x, y, z)
    NNS.dep(B)
    }
    @export
    """

    if np.any(np.isnan(x)):
        raise Exception("You have some missing values, please address.")
    if isinstance(x, pd.Series):
        x = x.to_frame()
    if isinstance(x, np.ndarray) and len(x.shape) == 1:
        x = x.reshape(-1, 1)  # series to matrix like

    if p_value:
        y_p = np.array([np.random.choice(range(1, 1 + y.shape[0]), y.shape[0], replace=False) for i in range(100)]).T
        x = np.insert(np.insert(y_p[:, 0:y_p.shape[1] - 2], 0, y, axis=1), 0, x, axis=1)
        y = None


    if y is not None:
        #x <- as.numeric(x)
        l = x.shape[0]

        #y <- as.numeric(y)
        obs = max(10, l/5)

        # Define segments
        if print_map:
            PART = NNS_part(
                x, y,
                order = None,
                obs_req = obs,
                min_obs_stop = True,
                type = "XONLY",
                Voronoi = True
            )
        else:
            PART = NNS_part(
                x, y,
                order = None,
                obs_req = obs,
                min_obs_stop = True,
                type = "XONLY",
                Voronoi = False
            )

        if (dim(PART$regression.points)[1]==0):
            return {
                "Correlation": 0,
                "Dependence" : 0
            }

        PART = PART['dt']
        PART = PART[~np.isnan(PART),:]
        # TODO
        PART[, weights := .N/l, by = prior.quadrant]
        weights = PART[, weights[1], by = prior.quadrant]$V1

        ll = expression(max(min(100, .N), 8))
        res = suppressWarnings(
            tryCatch(
                PART[,
                    sign(
                        cor(x[1:eval(ll)],y[1:eval(ll)])
                    )*summary(
                        lm(
                            y[1:eval(ll)]~
                            poly(x[1:eval(ll)], max(1, min(10, as.integer(sqrt(.N))-1)), raw = TRUE)
                        )
                    )$r.squared, by = prior.quadrant
                ],
                error = function(e) PART[, NNS_copula(cbind(x,y)), by = prior.quadrant]
            )
        )
        if (sum(is.na(res))>0)
            res[is.na(res)] = NNS_copula(cbind(x,y))

        # Compare each asymmetry
        res_xy = suppressWarnings(
            tryCatch(
                PART[,
                    sign(
                        cor(x[1:eval(ll)],(y[1:eval(ll)]))
                    )*summary(
                        lm(
                            abs(y[1:eval(ll)])~
                            poly(x[1:eval(ll)], max(1, min(10, as.integer(sqrt(.N))-1)), raw = TRUE)
                        )
                    )$r.squared, by = prior.quadrant
                ],
                error = function(e) PART[, NNS_copula(cbind(x,y)), by = prior.quadrant]
            )
        )
        res_yx = suppressWarnings(
            tryCatch(
                PART[,
                    sign(
                        cor(y[1:eval(ll)],(x[1:eval(ll)]))
                    )*summary(
                        lm(
                            abs(x[1:eval(ll)])~
                            poly(y[1:eval(ll)], max(1, min(10, as.integer(sqrt(.N))-1)), raw = TRUE)
                        )
                    )$r.squared, by = prior.quadrant
                ],
                error = function(e) PART[, NNS_copula(cbind(x,y)), by = prior.quadrant]
            )
        )

        if np.sum(np.isnan(res_xy))>0:
            res_xy[np.isnan(res_xy)] = NNS_copula(cbind(x,y))
        if np.sum(np.isnan(res_yx))>0:
            res_yx[np.isnan(res_yx)] = NNS_copula(cbind(x,y))

        if asym:
            dependence = sum(
                abs(res_xy$V1) * weights
            )
        else:
            dependence = np.max([
                sum(abs(res$V1) * weights),
                sum(abs(res_xy$V1) * weights),
                sum(abs(res_yx$V1) * weights)
            ])

        lx = PART[, length(unique(x))]
        ly = PART[, length(unique(y))]
        degree_x = min(10, max(1,lx-1), max(1,ly-1))

        I_x = lx > sqrt(l)
        I_y = ly > sqrt(l)
        I = I_x * I_y
        poly_base = dependence

        if I == 1:
            poly_base = suppressWarnings(
                tryCatch(
                    summary(
                        lm(abs(y)~poly(x, degree_x), raw = TRUE)
                    )$r.squared,
                    warning = function(w) dependence,
                    error = function(e) dependence
                )
            )
        dependence = mean(c(rep(dependence,3), poly_base))
        corr = np.mean([
            sum(res$V1 * weights),
            sum(res_xy$V1 * weights),
            sum(res_yx$V1 * weights)
        ])
        return {
            "Correlation": corr,
            "Dependence": dependence
        }
    elif p_value:
        # TODO: matplotlib
        nns.mc = apply(x, 2, function(g) NNS.dep(x[,1], g))

        ## Store results
        cors = unlist(lapply(nns.mc, "[[", 1))
        deps = unlist(lapply(nns.mc, "[[", 2))

        cor_lower_CI = LPM.VaR(.025, 0, cors[-c(1,2)])
        cor_upper_CI = UPM.VaR(.025, 0, cors[-c(1,2)])
        dep_lower_CI = LPM.VaR(.025, 0, deps[-c(1,2)])
        dep_upper_CI = UPM.VaR(.025, 0, deps[-c(1,2)])
        if print_map:
            # TODO: matplotlib
            original_par = par(no.readonly = TRUE)
            par(mfrow = c(1, 2))
            hist(cors[-c(1,2)], main = "NNS Correlation", xlab = NULL, xlim = c(min(cors), max(cors[-1])))
            abline(v = cors[2], col = "red", lwd = 2)
            mtext("Result", side = 3, col = "red", at = cors[2])
            abline(v =  cor_lower_CI, col = "red", lwd = 2, lty = 3)
            abline(v =  cor_upper_CI , col = "red", lwd = 2, lty = 3)
            hist(deps[-c(1,2)], main = "NNS Dependence", xlab = NULL, xlim = c(min(deps), max(deps[-1])))
            abline(v = deps[2], col = "red", lwd = 2)
            mtext("Result", side = 3, col = "red", at = deps[2])
            abline(v =  dep_lower_CI , col = "red", lwd = 2, lty = 3)
            abline(v =  dep_upper_CI , col = "red", lwd = 2, lty = 3)
            par(mfrow = original_par)

        return {
            "Correlation": as_numeric((cors)[2]),
            "Correlation p.value": min(
                LPM(0, cors[2], cors[-c(1,2)]),
                UPM(0, cors[2], cors[-c(1,2)])
            ),
            "Correlation 95% CIs": c(cor_lower_CI, cor_upper_CI),
            "Dependence": as_numeric((deps)[2]),
            "Dependence p.value": min(
                LPM(0, deps[2], deps[-c(1,2)]),
                UPM(0, deps[2], deps[-c(1,2)])
            ),
            "Dependence 95% CIs": c(dep_lower_CI, dep_upper_CI)
        }
    else:
        return NNS.dep.matrix(x, asym = asym)


__all__ = [
    "NNS_dep"
]